from anasymod.sim_ctrl.datatypes import *
from anasymod.enums import PortDir
from typing import Union
from anasymod.generators.codegen import CodeGenerator

io_obj_types = DigitalCtrlInput, DigitalCtrlOutput, DigitalSignal, AnalogCtrlInput, AnalogCtrlOutput, AnalogSignal, AnalogProbe
io_obj_types_union = Union[DigitalCtrlInput, DigitalCtrlOutput, DigitalSignal, AnalogCtrlInput, AnalogCtrlOutput, AnalogSignal, AnalogProbe]
io_obj_types_str_union = Union[DigitalCtrlInput, DigitalCtrlOutput, DigitalSignal, AnalogCtrlInput, AnalogCtrlOutput, AnalogSignal, AnalogProbe, str]

class GenAPI(CodeGenerator):
    """
    This is the general generator API custom tailored for generation of anasymod project files, e.g. top.
    """

    def __init__(self, line_ending='\n'):
        super().__init__(line_ending=line_ending)

    def gen_signal(self, io_obj: io_obj_types_union, default_value=None):
        """
        Generate a signal using the io object as input.
        :param io_obj: io object
        :param default_value: default value of signal, this works only for digital signals
        """
        raise NotImplementedError()

    def gen_port(self, io_obj: io_obj_types_union, direction: PortDir):
        """
        Generate an input port using the io_obj object as input.
        :param port: Port object
        """
        raise NotImplementedError()

    def gen_parameter(self, param_obj: io_obj_types_union):
        """
        Generate a parameter using the param_obj as input.

        :param param_obj: Object that includes a parameter
        """
        raise NotImplementedError()

    def gen_connection(self, io_objs: [io_obj_types_union]):
        """
        Connect an io_obj from upwards the hierarchy "upper_io_obj" to an io_obj further down the hierarchy "lower_io_obj".

        :param upper_io_obj:
        :param lower_io_obj:
        :return:
        """
        raise NotImplementedError()

    def assign_to(self, io_obj: io_obj_types_union, exp):
        """
        Assign 'exp' to signal 'io_obj.name'.
        NOTE: If the io_obj is of the AnalogSignal type, it is only possible to assign another signal of the same type
        or a constant value.

        :param io_obj: Signal to be assigned holds name of io_obj.name.
        :param exp: Expression to be assigned.
        """
        raise NotImplementedError()

    def decl_analog_port(self, io_obj: Union[AnalogCtrlInput, AnalogCtrlOutput]):
        """
        Generate a declaration for an analog signal. Necessary towork with SVREAL.

        :param io_obj:
        """
        raise NotImplementedError()

    def pass_analog_port_format(self, io_obj: io_obj_types_union, io_obj_format: io_obj_types_union):
        """
        Pass formatting information into instantiated module for the connected analog signal.

        :param io_connection: List of two IO objects that are connected to each other.
        """
        raise NotImplementedError()

class SVAPI(GenAPI):
    """
    This is the SV specific generator API custom for generation of anasymod project files, e.g. top.sv.
    """

    def __init__(self, line_ending='\n'):
        super().__init__(line_ending=line_ending)

    def gen_signal(self, io_obj: io_obj_types_union, default_value=None):
        """
        Generate a signal using the io object as input.
        :param io_obj: io object
        :param default_value: default value of signal, this works only for digital signals
        """
        if isinstance(io_obj, DigitalSignal):
            value = f' = {default_value}' if default_value is not None else ''
            width = f'[{str(io_obj.width)}  - 1:0] ' if io_obj.width > 1 else ''
            self.writeln(f"logic {width}{io_obj.name}{value};")
        else:
            if default_value is not None:
                raise Exception(f'ERROR: Default value provided for signal{io_obj.name}, which is an analog signal. '
                                f'Defaults can only be provided for digital signals!')
            if isinstance(io_obj, AnalogProbe):
                self.writeln(f"`MAKE_GENERIC_REAL({io_obj.name}, {io_obj.range}, {io_obj.width});")
            elif isinstance(io_obj, AnalogSignal):
                self.writeln(f"`MAKE_REAL({io_obj.name}, {io_obj.range});")

    def gen_port(self, io_obj: io_obj_types_union, direction: PortDir):
        """
        Generate an input port using the io_obj object as input.
        :param port: Port object
        :rtype str: generated SV for port.
        """

        if direction in [PortDir.IN]:
            if isinstance(io_obj, AnalogSignal):
                return f"`INPUT_REAL({io_obj.name})"
            elif isinstance(io_obj, (DigitalCtrlInput, DigitalCtrlOutput, DigitalSignal)):
                width = f'[{str(io_obj.width - 1)}:0] ' if io_obj.width > 1 else ''
                return f"input wire logic {'signed ' if io_obj.signed else ''}{width}{io_obj.name}"
        elif direction in [PortDir.OUT]:
            if isinstance(io_obj, AnalogSignal):
                return f"`OUTPUT_REAL({io_obj.name})"
            elif isinstance(io_obj, (DigitalCtrlInput, DigitalCtrlOutput, DigitalSignal)):
                width = f'[{str(io_obj.width - 1)}:0] ' if io_obj.width > 1 else ''
                return f"output wire logic {'signed ' if io_obj.signed else ''}{width}{io_obj.name}"
        elif direction in [PortDir.INOUT]:
            if isinstance(io_obj, AnalogSignal):
                raise Exception('Analog signals do not support the INOUT direction.')
            elif isinstance(io_obj, (DigitalCtrlInput, DigitalCtrlOutput, DigitalSignal)):
                width = f'[{str(io_obj.width - 1)}:0] ' if io_obj.width > 1 else ''
                return f"inout wire logic {'signed ' if io_obj.signed else ''}{width}{io_obj.name}"
        else:
            raise Exception(f"No valid direction provided: {direction}")

    def gen_parameter(self, param_obj):
        """
        Generate a parameter using the param_obj as input.

        :param param_obj: Object that includes a parameter
        :rtype str: returns generated parameter statement
        """

        return f"parameter {param_obj.type} {param_obj.name}={param_obj.value}"

    def gen_connection(self, io_objs: [io_obj_types_str_union]):
        """
        Connect an io_obj from upwards the hierarchy "upper_io_obj" to an io_obj further down the hierarchy "lower_io_obj".
        :rtype str: returns generated string for the connection
        """
        if len(io_objs) == 2:
            if isinstance(io_objs[1], (io_obj_types)): # Connect two signal type objects
                return f".{io_objs[0].name}({io_objs[1].name})"
            elif isinstance(io_objs[1], str): # Connect a signal type object with a fixed value
                return f".{io_objs[0].name}({io_objs[1]})"
            else:
                raise Exception(f'Wrong type passed for connection:{type(io_objs[1])}')
        else:
            raise Exception(
                f"Wrong number of io_objects provided in list, exactly two are requrend, provided were: {len(io_objs)}")

    def assign_to(self, io_obj: io_obj_types_str_union, exp):
        """
        Assign 'exp' to signal 'io_obj.name'.
        NOTE: If the io_obj is of the AnalogSignal type, it is only possible to assign another signal of the same type
        or a constant value.

        :param io_obj: Signal to be assigned holds name of io_obj.name.
        :param exp: Expression to be assigned.
        """
        if isinstance(io_obj, AnalogSignal) and not isinstance(io_obj, (AnalogCtrlInput, AnalogCtrlOutput, AnalogProbe)):
            if isinstance(exp, AnalogSignal):
                self.writeln(f"`ASSIGN_REAL({exp}, {io_obj.name});")
            if isinstance(exp, str):
                self.writeln(f"`ASSIGN_REAL({exp}, {io_obj.name});")
            else:
                try:
                    exp = str(float(exp))
                    self.writeln(f"`ASSIGN_CONST_REAL({exp}, {io_obj.name});")
                except:
                    raise Exception(f"The provided expression is not supported for Analog Signals, only assignment of another"
                                    f"AnalogSignal object or a constant value is supported; given: '{exp}'")
        elif isinstance(io_obj, (DigitalSignal, AnalogCtrlInput, AnalogCtrlOutput, AnalogProbe)):
            self.writeln(f"assign {io_obj.name} = {exp};")
        elif isinstance(io_obj, str):
            self.writeln(f"assign {io_obj} = {exp};")
        else:
            raise Exception(f'Not supported signal type provided:{type(io_obj)}')

    def decl_analog_port(self, io_obj: Union[AnalogCtrlInput, AnalogCtrlOutput, AnalogSignal]):
        """
        Generate a declaration for an analog signal. Necessary towork with SVREAL.

        :param io_obj:
        :rtype str: returns generated analog port
        """

        if isinstance(io_obj, (AnalogCtrlInput, AnalogCtrlOutput, AnalogSignal)):
            return f"`DECL_REAL({io_obj.name})"

    def pass_analog_port_format(self, io_obj: io_obj_types_union, io_obj_format: io_obj_types_union):
        """
        Pass formatting information into instantiated module for the connected analog signal.

        :param io_objs: List of two IO objects that are connected to each other.
        :rtype str: returns generated code for passing analog port format
        """

        if isinstance(io_obj, AnalogSignal):
            return f"`PASS_REAL({io_obj.name}, {io_obj_format.name})"
        else:
            raise Exception(f"ERROR: Wrong io type; supported only: "
                            f"{type(AnalogCtrlOutput), type(AnalogCtrlInput), type(AnalogSignal)}"
                            f"provided: {type(io_obj), type(io_obj_format)}")

class ModuleInst():
    """
    Container for Module Structure.
    """

    def __init__(self, api: GenAPI, name, inst_name=None):
        self.api = api

        self.name = name
        self.inst_name = inst_name if inst_name else f'{name}_i'
        self.analog_inputs = []
        self.analog_outputs = []
        self.digital_inputs = []
        self.digital_outputs = []
        self.digital_inouts = []
        self.parameters = []

        self.connections = []

    def add_inputs(self, io_objs: [io_obj_types_union], connections=None):
        if connections:
            if len(io_objs) != len(connections):
                raise Exception(f"ERROR: Each provided port object needs to have a connection counterpart! "
                                f"io_objs:{io_objs}, connections: {connections}")
            for io_obj, connection in zip(io_objs, connections):
                self.add_input(io_obj=io_obj, connection=connection)
        else:
            for io_obj in io_objs:
                self.add_input(io_obj=io_obj)

    def add_input(self, io_obj: io_obj_types_union, connection=None):
        if connection:
            if isinstance(connection, (AnalogSignal, DigitalSignal, str)):
                self.connect(io_obj=io_obj, io_obj_con=connection)
            else:
                raise Exception(f"Unsupported connection type was provided, supported types are. {io_obj_types_str_union}, "
                                f"provided: {type(connection)}")

        if isinstance(io_obj, (AnalogSignal)):
            self.analog_inputs.append(io_obj)
        else:
            self.digital_inputs.append(io_obj)

    def add_outputs(self, io_objs: [io_obj_types_union], connections=None):
        if connections:
            if len(io_objs) != len(connections):
                raise Exception(f"ERROR: Each provided port object needs to have a connection counterpart! "
                                f"io_objs:{io_objs}, connections: {connections}")
            for io_obj, connection in zip(io_objs, connections):
                self.add_output(io_obj=io_obj, connection=connection)
        else:
            for io_obj in io_objs:
                self.add_output(io_obj=io_obj)

    def add_output(self, io_obj: io_obj_types_union, connection=None):
        if connection:
            if isinstance(connection, (AnalogSignal, DigitalSignal, str)):
                self.connect(io_obj=io_obj, io_obj_con=connection)
            else:
                raise Exception(f"Unsupported connection type was provided, supported types "
                                f"are. {io_obj_types_union}, provided: {type(connection)}")

        if isinstance(io_obj, (AnalogSignal)):
            self.analog_outputs.append(io_obj)
        else:
            self.digital_outputs.append(io_obj)

    def add_inouts(self, io_objs: [io_obj_types_union], connections=None):
        if connections:
            if len(io_objs) != len(connections):
                raise Exception(f"ERROR: Each provided port object needs to have a connection counterpart! "
                                f"io_objs:{io_objs}, connections: {connections}")
            for io_obj, connection in zip(io_objs, connections):
                self.add_inout(io_obj=io_obj, connection=connection)
        else:
            for io_obj in io_objs:
                self.add_inout(io_obj=io_obj)

    def add_inout(self, io_obj: io_obj_types_union, connection=None):
        if connection:
            if isinstance(connection, (DigitalSignal, str)):
                self.connect(io_obj=io_obj, io_obj_con=connection)
            else:
                raise Exception(f"Unsupported connection type was provided, supported types "
                                f"are. {io_obj_types_union}, provided: {type(connection)}")

        if isinstance(io_obj, (AnalogSignal)):
            raise Exception('INOUT connect to AnalogSignal is not supported.')
        else:
            self.digital_inouts.append(io_obj)

    def add_parameters(self, io_objs: [io_obj_types_union]):
        for io_obj in io_objs:
            self.add_parameter(io_obj=io_obj)

    def add_parameter(self, io_obj: io_obj_types_union):
        self.parameters.append(io_obj)

    def connect(self, io_obj: io_obj_types_union, io_obj_con: io_obj_types_union):
        """
        Connect io_obj to io_obj_con. This is important for later module instantiations.

        :param io_obj:
        :param io_obj_con:
        """

        # ToDo: Improve type checkers to prevent multiple drivers.
        # Allow connections only of both signals are of type AnalogSignal, or both are of type DigitalSignal,
        # or signal is of type DigitalSignal and connecting signal is of type AnalogCtrlOutput/AnalogCtrlOutput; this is needed for
        # streaming Analog signals, or strings to account for constant values
        if (isinstance(io_obj, AnalogSignal) and isinstance(io_obj_con, AnalogSignal)) or \
                (isinstance(io_obj, DigitalSignal) and isinstance(io_obj_con, DigitalSignal)) or \
                (isinstance(io_obj, DigitalSignal) and isinstance(io_obj_con, (AnalogCtrlInput, AnalogCtrlOutput, AnalogProbe))) or \
                (isinstance(io_obj_con, str)):
            self.connections.append([io_obj, io_obj_con])
        else:
            raise Exception(f"IO types of objects to be connected do not match:"
                            f"io_obj: {type(io_obj)}, io_obj_con: {type(io_obj_con)}")

    def generate_header(self):
        """
        Generate full module header.
        """

        ### Check if module includes analog ports or parameters
        analog_ports = self.analog_inputs + self.analog_outputs

        if (analog_ports or self.parameters):
            self.api.writeln(f"module {self.name} #(")
            self.api.indent()
            ## Add parameter section
            if analog_ports:
                for analog_port in analog_ports[:-1]:
                    self.api.writeln(self.api.decl_analog_port(analog_port) + ",")
                self.api.writeln(self.api.decl_analog_port(analog_ports[-1]) + ("," if self.parameters else ""))
            if self.parameters:
                for parameter in self.parameters[:-1]:
                    self.api.writeln(self.api.gen_parameter(param_obj=parameter) + ",")
                self.api.writeln(self.api.gen_parameter(param_obj=self.parameters[-1]))
            self.api.dedent()
            self.api.writeln(f") (")
        else:
            self.api.writeln(f"module {self.name} (")

        #### Add port section

        io_lines = []
        for elem in (self.analog_inputs + self.digital_inputs):
            io_lines += [self.api.gen_port(io_obj=elem, direction=PortDir.IN)]
        for elem in (self.analog_outputs + self.digital_outputs):
            io_lines += [self.api.gen_port(io_obj=elem, direction=PortDir.OUT)]
        for elem in (self.digital_inouts):
            io_lines += [self.api.gen_port(io_obj=elem, direction=PortDir.INOUT)]

        self.api.indent()
        if len(io_lines) > 0:
            for io_line in io_lines[:-1]:
                self.api.writeln(f'{io_line},')
            self.api.writeln(f'{io_lines[-1]}')
        self.api.dedent()
        self.api.writeln(');')

    def generate_instantiation(self):
        """
        Generate instantiation for module.
        """

        ## Add parameter section
        analog_connections = []
        for connection in self.connections:
            # Remove analog control signals as for those signals no parameters will be transmitted
            if isinstance(connection[0], AnalogSignal):
                analog_connections.append(connection)

        if (analog_connections or self.parameters):
            self.api.indent()
            self.api.writeln(f"{self.name} #(")

            self.api.indent()
            if analog_connections:
                for analog_connection in analog_connections[:-1]:
                    self.api.writeln(self.api.pass_analog_port_format(io_obj=analog_connection[0], io_obj_format=analog_connection[1]) + ",")
                self.api.writeln(self.api.pass_analog_port_format(io_obj=analog_connections[-1][0], io_obj_format=analog_connections[-1][1]) + ("," if self.parameters else ""))

            if self.parameters:
                for parameter in self.parameters[:-1]:
                    self.api.writeln(self.api.gen_parameter(param_obj=parameter) + ",")
                self.api.writeln(self.api.gen_parameter(param_obj=self.parameters[-1]))

            self.api.dedent()
            self.api.writeln(f") {self.inst_name} (")
        else:
            self.api.indent()
            self.api.writeln(f"{self.name} {self.inst_name} (")

        ## Add port section
        self.api.indent()
        if self.connections:
            for connection in self.connections[:-1]:
                self.api.writeln(self.api.gen_connection(connection) + ",")
            self.api.writeln(self.api.gen_connection(self.connections[-1]))

        self.api.dedent()
        self.api.writeln(f");")


import os.path
import pathlib
import re
import multiprocessing
import shutil

from sys import platform
from glob import glob
from pathlib import Path
from anasymod.files import get_full_path, mkdir_p
from anasymod.util import vivado_search_key
from os import environ as env
from anasymod.enums import BoardNames
from anasymod.plugins import *
from anasymod.fpga_boards.boards import *
from anasymod.base_config import BaseConfig

class EmuConfig:
    def __init__(self, root, cfg_file, active_target, build_root=None):

        self._valid_ila_values = [1024, 2048, 4096, 8192, 16384, 32768, 65536, 131072]

        self.root = root

        # define and create build root
        self.build_root_base = build_root if build_root is not None else os.path.join(root, 'build')

        # define build root for functional models
        self.build_root_functional_models = os.path.join(self.build_root_base, 'models')

        self.build_root = os.path.join(self.build_root_base, active_target)
        if not os.path.exists(self.build_root):
            mkdir_p(self.build_root)

        self.cfg_file = cfg_file

        # Initialize config  dict
        self.cfg = Config(cfg_file=self.cfg_file)

        # Update config options by reading from config file
        self.cfg.update_config()

        # Try to initialize Inicio config_dict
        # TODO: add more reasonable defaults
        try:
            from inicio import config_dict
            self.cfg_dict = config_dict()
        except:
            self.cfg_dict = {
                'TOOLS_xilinx': '',
                'INICIO_TOOLS': '',
                'TOOLS_iverilog': '',
                'TOOLS_gtkwave': ''
            }

        # FPGA board configuration
        self.board = self._fetch_board()

        # Vivado configuration
        self.vivado_config = VivadoConfig(parent=self)

        # XSCT configuration
        self.xsct_config = XSCTConfig(parent=self)

        # GtkWave configuration
        self.gtkwave_config = GtkWaveConfig(parent=self)

        # SimVision configuration
        self.simvision_config = SimVisionConfig(parent=self)

        # Icarus configuration
        self.icarus_config = IcarusConfig(parent=self)

        # Xcelium configuration
        self.xcelium_config = XceliumConfig(parent=self)

    @property
    def ila_depth(self):
        if self.cfg.ila_depth in self._valid_ila_values:
            return self.cfg.ila_depth
        else:
            raise Exception(f'Provided value for ILA depth is not valid, allows values are:{self.cfg.ila_depth}, '
                            f'given: {self.cfg.ila_depth}')

    def _update_build_root(self, active_target):
        self.build_root = os.path.join(self.build_root_base, active_target)
        if not os.path.exists(self.build_root):
            mkdir_p(self.build_root)

    def _fetch_board(self):
        """
        Fetch FPGA board info.
        """

        if self.cfg.board_name == BoardNames.PYNQ_Z1:
            return PYNQ_Z1()
        elif self.cfg.board_name == BoardNames.ARTY_A7:
            return ARTY_A7()
        elif self.cfg.board_name == BoardNames.VC707:
            return VC707()
        elif self.cfg.board_name == BoardNames.ULTRA96:
            return ULTRA96()
        elif self.cfg.board_name == BoardNames.TE0720:
            return TE0720()
        elif self.cfg.board_name == BoardNames.ZC702:
            return ZC702()
        elif self.cfg.board_name == BoardNames.ZC706:
            return ZC706()
        elif self.cfg.board_name == BoardNames.ZCU102:
            return ZCU102()
        elif self.cfg.board_name == BoardNames.ZCU106:
            return ZCU106()
        elif self.cfg.board_name == BoardNames.ZCU111:
            return ZCU111()
        elif self.cfg.board_name == BoardNames.ARTY_200T_CUSTOM_LIDAR:
            return ARTY_200T_CUSTOM_LIDAR()
        else:
            raise Exception(f'The requested board {self.cfg.board_name} could not be found.')

class VivadoConfig():
    def __init__(self, parent: EmuConfig, vivado=None, version=None,
                 version_year=None, version_number=None):
        # save reference to parent config
        self.parent = parent

        # set project name
        self.project_name = 'prj'
        # intermediate variables for generic Xilinx path
        self.hints = []
        self.hints += [lambda: os.path.join(env['VIVADO_INSTALL_PATH'], 'bin')]
        if platform in {'win32', 'cygwin'}:
            try:
                xilinx_version_path = parent.cfg_dict['TOOLS_xilinx']
                xilinx_version = "20" + ".".join(xilinx_version_path.split(".")[0:2]).split("-")[1]
                self.hints += [
                    lambda: os.path.join(parent.cfg_dict['INICIO_TOOLS'], xilinx_version_path, "Vivado", xilinx_version, "bin")
                ]
            except:
                print('Skipping INICIO_TOOLS hint for Vivado.')
        # lsf options for tcl mode of Vivado
        self.lsf_opts_ls = ''
        self.lsf_opts = parent.cfg.lsf_opts
        if platform in {'linux', 'linux2'}:
            sorted_dirs = sorted(glob('/tools/Xilinx/Vivado/*.*'), key=vivado_search_key)
            self.hints.extend(lambda: os.path.join(dir_, 'bin') for dir_ in sorted_dirs)
            if 'CAMINO' in os.environ:
                self.lsf_opts = "-eh_ram 70000 -eh_ncpu 8 -eh_ui inicio_batch"
                self.lsf_opts_ls = "-eh_ram 70000 -eh_ncpu 8 -eh_dispatch LS_SHELL"

        self._vivado = vivado

        # version, year, number
        self._version = version
        self._version_year = version_year
        self._version_number = version_number

        # set various project options
        self.num_cores = multiprocessing.cpu_count()
        self.vio_name = 'vio_0'
        self.vio_inst_name = self.vio_name + '_i'
        self.ila_inst_name = 'u_ila_0'
        self.ila_reset = 'reset_probe'
        self.vio_reset = 'vio_i/rst'

    @property
    def vivado(self):
        if self._vivado is None:
            self._vivado = find_tool(name='vivado', hints=self.hints)
        return self._vivado

    @property
    def version(self):
        if self._version is None:
            self._version = pathlib.Path(self.vivado).parent.parent.name
        return self._version

    @property
    def version_year(self):
        if self._version_year is None:
            self._version_year = re.match(r'(\d+)\.(\d+)', self.version).groups()[0]
            self._version_year = int(self._version_year)
        return self._version_year

    @property
    def version_number(self):
        if self._version_number is None:
            self._version_number = re.match(r'(\d+)\.(\d+)', self.version).groups()[1]
            self._version_number = int(self._version_number)
        return self._version_number

class XSCTConfig():
    def __init__(self, parent: EmuConfig, xsct=None, version=None,
                 version_year=None, version_number=None, xsct_install_dir=None):
        # save reference to parent config
        self.parent = parent

        # set project name
        self.project_name = 'prj'
        # intermediate variables for generic Xilinx path
        self.hints = []
        self.hints += [lambda: os.path.join(env['XSCT_INSTALL_PATH'], 'bin')]
        if platform in {'win32', 'cygwin'}:
            try:
                xilinx_version_path = parent.cfg_dict['TOOLS_xilinx']
                xilinx_version = "20" + ".".join(xilinx_version_path.split(".")[0:2]).split("-")[1]
                self.hints += [
                    lambda: os.path.join(parent.cfg_dict['INICIO_TOOLS'], xilinx_version_path, "SDK", xilinx_version, "bin")
                ]
            except:
                print('Skipping INICIO_TOOLS hint for XSCT.')
        # lsf options for tcl mode of Vivado
        self.lsf_opts_ls = ''
        self.lsf_opts = parent.cfg.lsf_opts
        if platform in {'linux', 'linux2'}:
            sorted_dirs = sorted(glob('/tools/Xilinx/Vivado/*.*'), key=vivado_search_key)
            self.hints.extend(lambda: os.path.join(dir_, 'bin') for dir_ in sorted_dirs)
            if 'CAMINO' in os.environ:
                self.lsf_opts = "-eh_ram 70000 -eh_ncpu 8 -eh_ui inicio_batch"
                self.lsf_opts_ls = "-eh_ram 70000 -eh_ncpu 8 -eh_dispatch LS_SHELL"

        self._xsct = xsct
        self._xsct_install_dir = xsct_install_dir

        # version, year, number
        self._version = version
        self._version_year = version_year
        self._version_number = version_number

        # set various project options
        self.num_cores = multiprocessing.cpu_count()

    @property
    def xsct(self):
        if self._xsct is None:
            self._xsct = find_tool(name='xsct', hints=self.hints)
        return self._xsct

    @property
    def xsct_install_dir(self):
        if self._xsct_install_dir is None:
            self._xsct_install_dir = Path(self.xsct).resolve().parent.parent
        return self._xsct_install_dir

    @property
    def version(self):
        if self._version is None:
            self._version = pathlib.Path(self.xsct).parent.parent.name
        return self._version

    @property
    def version_year(self):
        if self._version_year is None:
            self._version_year = re.match(r'(\d+)\.(\d+)', self.version).groups()[0]
            self._version_year = int(self._version_year)
        return self._version_year

    @property
    def version_number(self):
        if self._version_number is None:
            self._version_number = re.match(r'(\d+)\.(\d+)', self.version).groups()[1]
            self._version_number = int(self._version_number)
        return self._version_number

class XceliumConfig():
    def __init__(self, parent: EmuConfig, xrun=None):
        # save reference to parent config
        self.parent = parent
        self.lsf_opts = parent.cfg.lsf_opts
        # set path to xrun binary
        self.hints = [lambda: os.path.join(env['XCELIUM_INSTALL_PATH'], 'bin')]
        self._xrun = xrun

        # name of TCL file
        self.tcl_input = 'input.tcl'



    @property
    def xrun(self):
        if self._xrun is None:
            try:
                self._xrun = find_tool(name='ifxxcelium', hints=self.hints)
                self.lsf_opts = "-eh_ncpu 4"
                #self._xrun += " execute"
            except:
                self._xrun = find_tool(name='xrun', hints=self.hints)
        return self._xrun

    @property
    def tcl_input_path(self):
        return os.path.join(self.parent.build_root, self.tcl_input)

class IcarusConfig():
    def __init__(self, parent: EmuConfig, iverilog=None, vvp=None):
        # save reference to parent config
        self.parent = parent

        # set path to iverilog and vvp binaries
        self.hints = [lambda: os.path.join(env['ICARUS_INSTALL_PATH'], 'bin'),
                      lambda: os.path.join(parent.cfg_dict['INICIO_TOOLS'], parent.cfg_dict['TOOLS_iverilog'], 'bin')]
        self._iverilog = iverilog
        self._vvp = vvp

        # name of output file
        self.output_file_name = 'a.out'

    @property
    def iverilog(self):
        if self._iverilog is None:
            self._iverilog = find_tool(name='iverilog', hints=self.hints)
        return self._iverilog

    @property
    def vvp(self):
        if self._vvp is None:
            self._vvp = find_tool(name='vvp', hints=self.hints)
        return self._vvp

    @property
    def output_file_path(self):
        return os.path.join(self.parent.build_root, self.output_file_name)

class GtkWaveConfig():
    def __init__(self, parent: EmuConfig, gtkwave=None):
        # save reference to parent config
        self.parent = parent

        # find gtkwave binary
        self.hints = [lambda: os.path.join(env['GTKWAVE_INSTALL_PATH'], 'bin'),
                      lambda: os.path.join(parent.cfg_dict['INICIO_TOOLS'], parent.cfg_dict['TOOLS_gtkwave'], 'bin')]
        self._gtkwave = gtkwave
        self.gtkw_config = None
        self.lsf_opts = parent.cfg.lsf_opts
        if 'CAMINO' in os.environ:
            self.lsf_opts = "-eh_ram 7000"

    @property
    def gtkwave(self):
        if self._gtkwave is None:
            self._gtkwave = find_tool(name='gtkwave', hints=self.hints)
        return self._gtkwave

class SimVisionConfig():
    def __init__(self, parent: EmuConfig, simvision=None):
        # save reference to parent config
        self.parent = parent

        # find simvision binary
        self.hints = [lambda: os.path.join(env['SIMVISION_INSTALL_PATH'], 'bin')]
        self._simvision = simvision
        self.svcf_config = None
        self.lsf_opts = parent.cfg.lsf_opts
        if 'CAMINO' in os.environ:
            self.lsf_opts = "-eh_ram 7000"

    @property
    def simvision(self):
        if self._simvision is None:
            self._simvision = find_tool(name='simvision', hints=self.hints)
        return self._simvision

class Config(BaseConfig):
    """
    Container to store all config attributes.
    """
    def __init__(self, cfg_file):
        super().__init__(cfg_file=cfg_file, section=ConfigSections.PROJECT)

        self.dec_bits = 24
        """ type(int) : number of bits used for decimation threshold value, which controls the ila capture feature """

        self.board_name = BoardNames.PYNQ_Z1
        """ type(str) : name of FPGA board, that shall be used for this project. """

        self.emu_clk_freq = 25e6
        """ type(float) : emulation frequency in Hz, that is used as main independent clk in the design. """

        self.preprocess_only = False
        """ type(bool) : indicate, whether or not only conduct elaboration and compile during simulation run. This will
            show all Macros expanded and currently only works for icarus verilog. """

        self.jtag_freq = 15e6
        """ type(float) : jtag communication frequency in Hz. This impacts the frequency used for programming the 
            bitstream onto the FPGA, a value that is set too high might cause stability issues. """

        self.plugins = ['msdsl']
        """ type(list(str)) : list of plugins that shall be used in the scope of this project. """

        self.dt = 0.1e-6
        """ type(float) : globally used dt value for projects, that strictly use fixed-timestep-based models. """

        self.dt_scale = 1e-15
        """ type(float) : resolution for dt request and response signals (e.g., "1e-15" means 1fs resolution). """

        self.dt_width = 32
        """ type(int) : number of bits used for dt request and response signals """

        self.time_width = 64
        """ type(int) : number of bits used for emulation time signal.  TODO: explore issue with using
            values larger than 39 (related to ILA). """
        
        self.lsf_opts = ''
        """ type(string) : LSF options which can be passed to our tool wrapper commands in the Inicio Flowpackage
            Execution handler options, e.g. '-eh_ncpu 8' can be used"""

        self.ila_depth = 4096
        """ type(int) : number of samples that can be stored in the ILA for each signal that is probed. Valid values are:
            1,024, 2,048, 4,096, 8,192, 16,384, 32,768, 65,536, 131,072. Also when choosing this value keep in mind,
            that this will consume Block Memory. Possible values are further limited by the number of signals to be 
            stored, number of block memory cells available on the FPGA board and also block memory that was consumed 
            already be the design itself. """

        self.cpu_debug_mode = False
        """ type(bool) : Enables probing all signals for either everything below the testbench, or for modules 
            provided in the cpu_debug_hierarchies attribute."""

        self.cpu_debug_hierarchies = None
        """ type(str, list) : Either tuple of depth and absolute path or list of tuples of depth and absolute path to 
            design modules, for which all signals shall be stored in result file. e.g.:
            [(0, top.tb_i.filter_i)]"""

        self.flatten_hierarchy = 'rebuilt'
        """ type(str) : Flattening strategy used in synthesis by Vivado.  Can be 'none', 'full', or 'rebuilt'.
            Choosing 'none' is a good strategy for debugging synthesis issues, while 'full' may allow for
            better synthesis results.  'rebuilt' is a trade-off between readability and performance, but
            sometimes has bugs, particularly when more complex Verilog features like hierarchical references and
            interfaces are used."""

        self.vivado_stack = None
        """ type(int) : Stack size used by Vivado.  If you are getting an error during synthesis that
            looks like "Segmentation fault "$RDI_PROG"", try setting this to 2000 or higher
            (https://www.xilinx.com/support/answers/64434.html)."""

        self.no_time_manager = False
        """ type(bool) : If True, do not instantiate the emu_clk generator, default oscillator,
            time manager, time control signals, and instead simply wire emu_clk_2x directly
            to emu_clk.  This can help the design run at higher frequencies, if no time management
            is required. """

        self.treat_v_as_sv = False
        """ type(bool) : If True, treat Verilog (*.v) files as SystemVerilog (*.sv) in Vivado.
            This option requires there to be at least one *.v file in the project sources. """

def find_tool(name, hints=None, sys_path_hint=True):
    # set defaults
    if hints is None:
        hints = []



    # add system path as the last "hint" if desired (default behavior)
    if sys_path_hint:
        hints.append(lambda: None)

    # first check the system path for the tool
    tool_path = shutil.which(name)

    # if the tool isn't found in the system path, then try out the hints in order
    if tool_path is None:
        for hint in hints:
            try:
                if callable(hint):
                    tool_path = shutil.which(name, path=hint())
                else:
                    tool_path = shutil.which(name, path=hint)
            except:
                continue

            if tool_path is not None:
                break

    if tool_path is not None:
        return get_full_path(tool_path)
    else:
        raise KeyError(f'Tool:{name} could not be found')





from anasymod.templates.generic_ip import TemplGenericIp
from anasymod.config import EmuConfig
from anasymod.structures.structure_config import StructureConfig
from anasymod.sim_ctrl.datatypes import DigitalCtrlInput, DigitalCtrlOutput, DigitalSignal, AnalogSignal, AnalogCtrlInput, AnalogCtrlOutput

class TemplVIO(TemplGenericIp):
    def __init__(self, scfg: StructureConfig, ip_dir):

        ####################################################
        # Generate ip core config for vio wizard
        ####################################################

        crtl_inputs = scfg.analog_ctrl_inputs + scfg.digital_ctrl_inputs
        ctrl_outputs = scfg.analog_ctrl_outputs + scfg.digital_ctrl_outputs

        props = {}

        # handle input ports
        for k, input in enumerate(ctrl_outputs):
            # check that the signal type can be handled
            if not isinstance(input, (AnalogCtrlOutput, DigitalCtrlOutput)):
                raise Exception(f"Provided signal type:{type(input)} is not supported!")
            # set signal width
            props[f'CONFIG.C_PROBE_IN{k+0}_WIDTH'] = str(input.width)

        # handle output ports
        for k, output in enumerate(crtl_inputs):
            # check that the signal type can be handled
            if not isinstance(output, (AnalogCtrlInput, DigitalCtrlInput)):
                raise Exception(f"Provided signal type:{type(output)} is not supported!")
            # set signal width
            props[f'CONFIG.C_PROBE_OUT{k + 0}_WIDTH'] = str(output.width)
            # set initial value of probe
            if output.init_value is not None:
                init_value = output.init_value
                if isinstance(output, AnalogCtrlInput):
                    init_value = output.float_to_fixed(init_value)
                props[f'CONFIG.C_PROBE_OUT{k+0}_INIT_VAL'] = str(init_value)

        props['CONFIG.C_NUM_PROBE_IN'] = str(len(ctrl_outputs))
        props['CONFIG.C_NUM_PROBE_OUT'] = str(len(crtl_inputs))

        ####################################################
        # Prepare Template substitutions
        ####################################################

        super().__init__(ip_name='vio', ip_dir=ip_dir, props=props)

def main():
    from anasymod.targets import FPGATarget
    print(TemplVIO(target=FPGATarget(prj_cfg=EmuConfig(root='test', cfg_file=''))).render())

if __name__ == "__main__":
    main()


import os

from anasymod.defines import Define
from anasymod.config import EmuConfig
from anasymod.base_config import BaseConfig
from anasymod.enums import ConfigSections, FPGASimCtrl, ResultFileTypes
from anasymod.structures.structure_config import StructureConfig
from anasymod.structures.module_top import ModuleTop
from anasymod.structures.module_clk_manager import ModuleClkManager
from anasymod.sources import VerilogSource, VerilogHeader, FirmwareFile
from anasymod.sim_ctrl.uart_zynq_ctrlinfra import UARTControlInfrastructure
from anasymod.sim_ctrl.vio_ctrlinfra import VIOControlInfrastructure
from anasymod.structures.module_traceport import ModuleTracePort
from anasymod.structures.module_emu_clks import ModuleEmuClks
from anasymod.structures.module_time_manager import ModuleTimeManager
from anasymod.sim_ctrl.vio_ctrlapi import VIOCtrlApi
from anasymod.sim_ctrl.uart_zynq_ctrlapi import UARTCtrlApi
from anasymod.files import anasymod_root, anasymod_header
from anasymod.util import expand_searchpaths

from anasymod.structures.module_viosimctrl import ModuleVIOSimCtrl

class Target():
    """
    This class inherits all source and define objects necessary in order to run actions for a specific target.
    """
    def __init__(self, prj_cfg: EmuConfig, plugins: list, name, target_type, float_type):
        self.prj_cfg = prj_cfg
        self.plugins = plugins
        self.float_type = float_type

        # Instantiate Simulation ControlInfrastructure Interface
        self.ctrl = None
        """ :type : ControlInfrastructure"""

        self._name = name

        # Initialize Probe Objs; there can be a probe obj for each supported simulator and synthesizer
        self.probes = {}
        """ :type : dict(Probe)"""

        # Initialize content dict  to store design sources and defines
        self.content = Content()

        # Initialize target_config
        self.cfg = Config(cfg_file=self.prj_cfg.cfg_file, prj_cfg=self.prj_cfg, name=self._name, target_type=target_type)

        # Update config options by reading from config file
        self.cfg.update_config(subsection=self._name)

        # Initialize structure configuration
        self.str_cfg = StructureConfig(prj_cfg=self.prj_cfg, simctrl_path= self.expanded_simctrl_path)

    def set_tstop(self):
        """
        Add define statement to specify tstop
        """
        if self.cfg.tstop is not None:
            self.content.defines.append(Define(name='TSTOP_MSDSL', value=self.cfg.tstop))

    def assign_fileset(self, fileset: dict):
        """
        Add a fileset to the target by adding defines and sources to the according attributes

        :param fileset: fileset that shall be added to target
        :type fileset: dict
        """

        for k in fileset.keys():
            setattr(self.content, k, getattr(self.content, k) + fileset[k])

    def update_structure_config(self):
        self.str_cfg.cfg.update_config(subsection=self._name)

    def gen_structure(self):
        """
        Generate toplevel, IPCore wrappers, debug infrastructure and clk manager.
        """

        # Add anasymod header
        self.content.verilog_headers += [VerilogHeader(str(anasymod_header()), name='anasymod_header')]

        # Generate toplevel and add to target sources
        toplevel_path = os.path.join(self.prj_cfg.build_root, 'gen_top.sv')
        with (open(toplevel_path, 'w')) as top_file:
            top_file.write(ModuleTop(target=self).render())
        self.content.verilog_sources += [VerilogSource(files=toplevel_path, name='top')]

        # Build control structure and add all sources to project
        if self.ctrl is not None:
            self.ctrl.gen_ctrlwrapper(str_cfg=self.str_cfg, content=self.content)
        else:
            #ToDO: needs to be cleaned up, should have individual module for pc simulation control
            with (open(os.path.join(self.prj_cfg.build_root, 'gen_ctrlwrap.sv'), 'w')) as ctrl_file:
                ctrl_file.write(ModuleVIOSimCtrl(scfg=self.str_cfg, plugin_includes=self.plugins).render())
            self.content.verilog_sources += [VerilogSource(files=os.path.join(self.prj_cfg.build_root, 'gen_ctrlwrap.sv'),
                                                           name='gen_ctrlwrap')]

        # Include the source code for an oscillator if needed
        if self.str_cfg.use_default_oscillator:
            osc_model_anasymod = anasymod_root() / 'verilog' / 'osc_model_anasymod.sv'
            self.content.verilog_sources += [VerilogSource(files=str(osc_model_anasymod), name='osc_model_anasymod')]

        # Include the source code for the anasymod control block
        ctrl_anasymod = anasymod_root() / 'verilog' / 'ctrl_anasymod.sv'
        self.content.verilog_sources += [VerilogSource(files=str(ctrl_anasymod), name='ctrl_anasymod')]

        # Generate clk management wrapper and add to target sources
        clkmanagerwrapper_path = os.path.join(self.prj_cfg.build_root, 'gen_clkmanager_wrap.sv')
        with (open(clkmanagerwrapper_path, 'w')) as clkm_file:
            clkm_file.write(ModuleClkManager(scfg=self.str_cfg).render())
        self.content.verilog_sources += [VerilogSource(files=clkmanagerwrapper_path,
                                                       name='gen_clkmanager_wrap')]

        # Generate traceport wrapper and add to target sources
        trapwrapper_path = os.path.join(self.prj_cfg.build_root, 'gen_traceport_wrap.sv')
        with (open(trapwrapper_path, 'w')) as trap_file:
            trap_file.write(ModuleTracePort(scfg=self.str_cfg).render())
        self.content.verilog_sources += [VerilogSource(files=trapwrapper_path,
                                                       name='gen_traceport_wrap')]

        # Generate emulation clk gen module and add to target sources
        gen_emu_clks_path = os.path.join(self.prj_cfg.build_root, 'gen_emu_clks.sv')
        with (open(gen_emu_clks_path, 'w')) as emu_clks_file:
            emu_clks_file.write(ModuleEmuClks(scfg=self.str_cfg, pcfg=self.prj_cfg).render())
        self.content.verilog_sources += [VerilogSource(files=gen_emu_clks_path,
                                                       name='gen_emu_clks')]

        # Generate time manager and add to target sources
        timemanager_path = os.path.join(self.prj_cfg.build_root, 'gen_time_manager.sv')
        with (open(timemanager_path, 'w')) as timemanager_file:
            timemanager_file.write(ModuleTimeManager(scfg=self.str_cfg, pcfg=self.prj_cfg, plugin_includes=self.plugins).render())
        self.content.verilog_sources += [VerilogSource(files=timemanager_path,
                                                       name='gen_time_manager')]

    @property
    def project_root(self):
        return os.path.join(self.prj_cfg.build_root, self.prj_cfg.vivado_config.project_name)

    @property
    def result_name_raw(self):
        return f"{self.cfg.top_module}_{self._name}.{self.cfg.result_type_raw}"

    @property
    def result_path_raw(self):
        return os.path.join(self.prj_cfg.build_root, r"raw_results", self.result_name_raw)

    @property
    def expanded_simctrl_path(self):
        """
        Path used to load simctrl file. This allows to use a custom simctrl configuration for a target.

        :return: str
        """
        return "".join(expand_searchpaths(paths=self.cfg.simctrl_path, rel_path_reference=self.prj_cfg.root))

class CPUTarget(Target):
    def __init__(self, prj_cfg: EmuConfig, plugins: list, float_type:bool, name=r"sim"):
        target_type = ConfigSections.CPU_TARGET
        super().__init__(prj_cfg=prj_cfg, plugins=plugins, name=name, target_type=target_type, float_type=float_type)

        self.cfg.result_type_raw = ResultFileTypes.VCD

class FPGATarget(Target):
    def __init__(self, prj_cfg: EmuConfig, plugins: list,  float_type:bool, name=r"fpga"):
        target_type = ConfigSections.FPGA_TARGET
        super().__init__(prj_cfg=prj_cfg, plugins=plugins, name=name, target_type=target_type, float_type=float_type)

        self.cfg.result_type_raw = ResultFileTypes.CSV

        # use a different default TSTOP value, which should provide about 0.1 ps timing resolution and plenty of
        # emulation time for most purposes.
        self._ip_cores = []

    def gen_structure(self):
        """
        Generate toplevel, IPCore wrappers, debug/ctrl infrastructure for FPGA and clk manager.
        """
        super().gen_structure()
        self.ctrl.gen_ctrl_infrastructure(content=self.content)

    def setup_ctrl_ifc(self, debug=False):
        """
        Setup the control interface according to what was provided in the project configuration. default is VIVADO_VIO
        mode, which does not possess a direct control interface via anasymod.

        :rtype: ControlInfrastructure
        """

        if self.cfg.fpga_sim_ctrl == FPGASimCtrl.VIVADO_VIO:
            print("No direct control interface from anasymod selected, Vivado VIO interface enabled.")
            self.ctrl = VIOControlInfrastructure(prj_cfg=self.prj_cfg, plugin_includes=self.plugins)
            self.ctrl_api = VIOCtrlApi(result_path=self.cfg.vcd_path, result_path_raw=self.result_path_raw,
                                       result_type_raw=self.cfg.result_type_raw, pcfg=self.prj_cfg, scfg=self.str_cfg,
                                       bitfile_path=self.bitfile_path, ltxfile_path=self.ltxfile_path,
                                       float_type=self.float_type, debug=debug)
        elif self.cfg.fpga_sim_ctrl == FPGASimCtrl.UART_ZYNQ:
            print("Direct anasymod FPGA simulation control via UART enabled.")
            self.ctrl = UARTControlInfrastructure(prj_cfg=self.prj_cfg,
                                                  scfg=self.str_cfg,
                                                  plugin_includes=self.plugins,
                                                  tcfg= self.cfg
                                                  )
            self.ctrl_api = UARTCtrlApi(result_path=self.cfg.vcd_path, result_path_raw=self.result_path_raw,
                                        result_type_raw=self.cfg.result_type_raw, prj_cfg=self.prj_cfg,
                                        scfg=self.str_cfg, content=self.content, ltxfile_path=self.ltxfile_path,
                                        top_module=self.cfg.top_module, project_root=self.project_root,
                                        float_type=self.float_type
                                        )
        else:
            raise Exception("ERROR: No FPGA simulation control was selected, shutting down.")


    @property
    def probe_cfg_path(self):
        return os.path.join(self.project_root, 'probe_config.txt')

    @property
    def bitfile_path(self):
        return os.path.join(self.project_root, f'{self.prj_cfg.vivado_config.project_name}.runs', 'impl_1',
                            f"{self.cfg.top_module}.bit")

    @property
    def ltxfile_path(self):
        return os.path.join(self.project_root, f'{self.prj_cfg.vivado_config.project_name}.runs', 'impl_1',
                            f"{self.cfg.top_module}.ltx")

    @property
    def ip_dir(self):
        return os.path.join(self.project_root, f'{self.prj_cfg.vivado_config.project_name}.srcs', 'sources_1', 'ip')

class Config(BaseConfig):
    """
    Container to store all config attributes.
    """
    def __init__(self, cfg_file, prj_cfg, name, target_type):
        super().__init__(cfg_file=cfg_file, section=target_type)
        self.tstop = 10.0
        """ type(float) : simulation duration for target in seconds. """

        self.emu_clk_freq =25e6
        """ type(float) : emulation frequency in Hz, that is used as main independent clk in the design. """

        self.top_module = 'top'
        """ type(str) : name ot the top-level module, that is used for this target. """

        self.custom_top = False
        """ type(bool) : indicates, whether or not a custom top is used. """

        self.custom_zynq_firmware = False
        """ type(bool) : indicates, whether or not a custom firmware for zynq PS to control USB_UART is used. """

        self.vcd_name = f"{self.top_module}_{name}.vcd"
        """ type(str) : name of the converted vcd simulation result file. """

        self.vcd_path = os.path.join(prj_cfg.build_root, r"vcd", self.vcd_name)
        """ type(str) : path used to store converted vcd simulation result file. """
        # TODO: move these paths to toolchain specific config, which shall be instantiated in the target class

        self.result_type_raw = None
        """ type(float) : simulation result format used for the selected target. """

        self.fpga_sim_ctrl = FPGASimCtrl.VIVADO_VIO
        """ type(float) : FPGA simulation control interface used for this target. """

        self.simctrl_path = os.path.join(prj_cfg.root, 'simctrl.yaml')
        """ type(str) : Path used to load simctrl file. This allows to use a custom simctrl configuration for a target. """

class Content():
    """
    Container storing all Project specific source files, such as RTL, eSW code, blockdiagrams, IP core configs,
    constraint files, ... .
    """

    def __init__(self):
        self.verilog_sources = []
        """:type : List[VerilogSource]"""
        self.verilog_headers = []
        """:type : List[VerilogHeader]"""
        self.vhdl_sources = []
        """:type : List[VHDLSource]"""
        self.include_dirs = []
        """:type : List[IncludeDir]"""
        self.defines = []
        """:type : List[Define]"""
        self.edif_files = []
        """:type : List[EDIFFile]"""
        self.firmware_files = []
        """:type : List[FirmwareFile]"""
        self.xci_files = []
        """:type : List[XCIFile]"""
        self.tcl_files = []
        """:type : List[TCLFile]"""
        self.xdc_files = []
        """:type : List[XDCFile]"""
        self.mem_files = []
        """:type : List[MEMFile]"""
        self.bd_files = []
        """:type : List[BDFile]"""
        self.ip_repos = []
        """:type : List[IPRepo]"""
        self.functional_models = []
        """:type : List[FunctionalModel]"""



import os.path

from anasymod.templates.templ import JinjaTempl
from anasymod.util import back2fwd


class TemplGenericIp(JinjaTempl):
    def __init__(self, ip_name, ip_module_name=None, props=None, ip_dir=None):
        super().__init__()
        # set defaults
        if props is None:
            props = {}

        if ip_module_name is None:
            ip_module_name = ip_name + '_0'

        self.ip_name = ip_name
        self.ip_module_name = ip_module_name
        self.ip_xci_path = back2fwd(os.path.join(ip_dir, ip_module_name, f'{ip_module_name}.xci'))

        self.props = props

    TEMPLATE_TEXT = '''
# start auto-generated code for {{subst.ip_name}}

create_ip -name {{subst.ip_name}} -vendor xilinx.com -library ip -module_name {{subst.ip_module_name}}

set_property -dict [list \\
{%- for propname, propvalue in subst.props.items() %}
    {{propname}} {{'{'}}{{ propvalue }}{{'}'}} \\
{%- endfor %}
] [get_ips {{subst.ip_module_name}}]

generate_target instantiation_template [get_files "{{subst.ip_xci_path}}"]

# end auto-generated code for {{subst.ip_name}}

'''

def main():
    print(TemplGenericIp(ip_name='ip_name', ip_module_name='ip_module_name', ip_dir='ip_dir').render())

if __name__ == "__main__":
    main()


import shutil
import re

from pathlib import Path

from anasymod.generators.xsct import XSCTTCLGenerator

from anasymod.templates.xsct_build import TemplXSCTBuild
from anasymod.templates.xsct_program import TemplXSCTProgram
from anasymod.config import EmuConfig

class XSCTEmulation(XSCTTCLGenerator):
    """
    Generate and execute Vivado TCL scripts to generate a bitstream, run an emulation of FPGA for non-interactive mode,
    or launch an FPGA emulation for interactive mode and pass the handle for interactive control.
    """

    def __init__(self, pcfg: EmuConfig, top_module, project_root, content):
        super().__init__(pcfg=pcfg)
        self.top_module = top_module
        self.project_root = project_root
        self.content = content

        self.err_str = re.compile('^(Failed)')

    @property
    def impl_dir(self):
        return (
            Path(self.project_root) /
            f'{self.pcfg.vivado_config.project_name}.runs' /
            'impl_1'
        )

    @property
    def bit_path(self):
        return self.impl_dir / f'{self.top_module}.bit'

    @property
    def tcl_path(self):
        if self.pcfg.board.is_ultrascale:
            return self.impl_dir / 'psu_init.tcl'
        else:
            return self.impl_dir / 'ps7_init.tcl'

    @property
    def hw_path(self):
        if self.pcfg.xsct_config.version_year < 2020:
            return self.impl_dir / f'{self.top_module}.sysdef'
        else:
            return self.impl_dir / f'{self.top_module}.xsa'

    def build(self, create=True, copy_files=True, build=True):
        # determine SDK path
        sdk_path = (Path(self.project_root) /
                    f'{self.pcfg.vivado_config.project_name}.sdk')

        # clear the SDK directory
        if create:
            shutil.rmtree(sdk_path, ignore_errors=True)
            sdk_path.mkdir(exist_ok=True, parents=True)

        # copy over the firmware sources
        if copy_files:
            src_path = sdk_path / 'sw' / 'src'
            src_path.mkdir(exist_ok=True, parents=True)
            for src in self.content.firmware_files:
                if src.files is not None:
                    for file_ in src.files:
                        shutil.copy(str(file_), str(src_path / Path(file_).name))

        # determine the processor name
        if self.pcfg.board.is_ultrascale:
            proc_name = 'psu_cortexa53_0'
        else:
            proc_name = 'ps7_cortexa9_0'

        # generate the build script
        self.write(
            TemplXSCTBuild(
                sdk_path=sdk_path,
                hw_path=self.hw_path,
                version_year=self.pcfg.xsct_config.version_year,
                version_number=self.pcfg.xsct_config.version_number,
                proc_name=proc_name,
                create=create,
                build=build
            ).text
        )

        # run the build script while checking for errors
        err_str = re.compile(r'(: error:)|(make.*: \*\*\* .* Error \d+)')
        self.run('sdk.tcl', err_str=err_str)

    def program(self, **kwargs):
        # determine SDK path
        sdk_path = (Path(self.project_root) /
                    f'{self.pcfg.vivado_config.project_name}.sdk')

        # generate the build script
        self.write(
            TemplXSCTProgram(
                sdk_path=sdk_path,
                bit_path=self.bit_path,
                hw_path=self.hw_path,
                tcl_path=self.tcl_path,
                is_ultrascale=self.pcfg.board.is_ultrascale,
                xsct_install_dir=self.pcfg.xsct_config.xsct_install_dir,
                no_rev_check=self.pcfg.board.no_rev_check,
                pcfg=self.pcfg,
                **kwargs
            ).text
        )

        # run the programming script. xsct appears to return an error code
        # when the FPGA board is not plugged in or is not powered up, so
        # parsing for errors doesn't appear to be necessary in this case)
        self.run('program.tcl', err_str=self.err_str)



from jinja2 import Environment

class JinjaTempl:
    def __init__(self, trim_blocks=False, lstrip_blocks=False):
        self.trim_blocks = trim_blocks
        self.lstrip_blocks = lstrip_blocks

    def render(self):
        t = Environment(trim_blocks=self.trim_blocks, lstrip_blocks=self.lstrip_blocks).from_string(self.TEMPLATE_TEXT)
        #t = Template(self.TEMPLATE_TEXT)

        return t.render(subst=self)+'\n'

    TEMPLATE_TEXT = ''


from anasymod.templates.generic_ip import TemplGenericIp
from anasymod.targets import FPGATarget
from anasymod.config import EmuConfig

class TemplClkWiz(TemplGenericIp):
    def __init__(self, target: FPGATarget):
        # defaults
        scfg = target.str_cfg
        pcfg = target.prj_cfg

        ####################################################
        # Generate ip core config for clk wizard
        ####################################################
        props = {}
        # Add input clks
        props['CONFIG.PRIM_IN_FREQ'] = str(pcfg.board.clk_freq * 1e-6)
        if len(scfg.clk_i) == 2:
            props['CONFIG.PRIM_SOURCE'] = 'Differential_clock_capable_pin'
        elif len(scfg.clk_i) == 1:
            props['CONFIG.PRIM_SOURCE'] = 'Single_ended_clock_capable_pin'
        else:
            raise Exception("Wrong number of master clk pins is provided")

        # Add emu_clk_2x
        props[f'CONFIG.CLKOUT1_USED'] = 'true'
        props['CONFIG.CLKOUT1_REQUESTED_OUT_FREQ'] = (scfg.emu_clk_2x.freq * 1e-6)

        # Add dbg_clk
        props[f'CONFIG.CLKOUT2_USED'] = 'true'
        props[f'CONFIG.CLKOUT2_REQUESTED_OUT_FREQ'] = (scfg.dbg_clk.freq * 1e-6)

        # Add additional independent clks
        for k, clk in enumerate(scfg.clk_independent):
            props[f'CONFIG.CLKOUT{k + 3}_USED'] = 'true'
            props[f'CONFIG.CLKOUT{k + 3}_REQUESTED_OUT_FREQ'] = (clk.freq * 1e-6)

        ####################################################
        # Prepare Template substitutions
        ####################################################

        props[f'CONFIG.NUM_OUT_CLKS'] = len(scfg.clk_independent) + 2

        super().__init__(ip_name='clk_wiz', ip_dir=target.ip_dir, props=props)

def main():
    print(TemplClkWiz(target=FPGATarget(prj_cfg=EmuConfig(root='test', cfg_file=''))).render())

if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod
from anasymod.targets import CPUTarget
from anasymod.config import EmuConfig

class Simulator(ABC):
    def __init__(self, target: CPUTarget, flags=None):
        # set defaults
        if flags is None:
            flags = []

        # save settings
        self.cfg = target.prj_cfg
        self.target = target
        self.flags = flags

    @abstractmethod
    def simulate(self):
        pass



from anasymod.structures.signal_base import Signal
from anasymod.enums import PortDir

class Port():
    """
    This is the base class for port generation.
    """
    def __init__(self, name, width, direction=None, **kwargs):
        self.name = name
        self.width = width
        self.connection = None
        self.direction = direction
        self.__init_value = kwargs['init_value'] if 'init_value' in kwargs else None

    @property
    def init_value(self):
        if self.direction == PortDir.OUT:
            return self.__init_value

    def connect(self, signal: Signal):
        """
        Connects a signal to the port.
        """
        self.connection = signal.name

class PortIN(Port):
    def __init__(self, name, width=1):
        super().__init__(name=name, width=width, direction=PortDir.IN)

class PortOUT(Port):
    def __init__(self, name, width=1, direction=PortDir.OUT, init_value=None):
        super().__init__(name=name, width=width)


from abc import ABC, abstractmethod

from anasymod.config import EmuConfig
from anasymod.targets import Target

class Viewer(ABC):
    def __init__(self, target: Target):
        self.cfg = target.prj_cfg
        self.target = target

    @abstractmethod
    def view(self, result_file=None):
        pass


import os
from anasymod.sim_ctrl.ctrlinfra import ControlInfrastructure
from anasymod.structures.module_uartsimctrl import ModuleUARTSimCtrl
from anasymod.structures.module_regmapsimctrl import ModuleRegMapSimCtrl
from anasymod.sources import VerilogSource, BDFile, FirmwareFile
from anasymod.files import get_from_anasymod
from anasymod.structures.structure_config import StructureConfig
from anasymod.structures.firmware_gpio import FirmwareGPIO
from anasymod.structures.uart_zynq_firmware_appcode import UartZynqFirmwareAppCode
#from anasymod.targets import Config as tcfg

class UARTControlInfrastructure(ControlInfrastructure):
    def __init__(self, prj_cfg, scfg: StructureConfig, plugin_includes, tcfg):
        super().__init__(prj_cfg=prj_cfg, plugin_includes=plugin_includes)
        self.scfg = scfg
        self.tcfg = tcfg

        # Initialize internal variables
        self._simctrlregmap_path = os.path.join(prj_cfg.build_root, 'gen_ctrlregmap.sv')

        #TODO: add path to elf file and add structure for storing ctrl ifc dependent files,
        # also including the BD; later test if .tcl script for creating BD would be
        # beneficial/at least there should be a script for creating the .bd file for a
        # new Vivado version, also add binary path to xsct interface

    def gen_ctrlwrapper(self, str_cfg: StructureConfig, content):
        """
        Generate RTL design for control infrastructure. This will generate the register map, add the
        block diagram including the zynq PS and add the firmware running on the zynq PS.
        """

        # Generate simulation control wrapper and add to target sources
        with (open(self._simctrlwrap_path, 'w')) as ctrl_file:
           ctrl_file.write(ModuleUARTSimCtrl(scfg=self.scfg).render())

        content.verilog_sources += [VerilogSource(files=self._simctrlwrap_path, name='simctrlwrap')]

    def gen_ctrl_infrastructure(self, content):
        """
        Generate RTL design for FPGA specific control infrastructure, depending on the interface
        selected for communication. For UART_ZYNQ control a register map, ZYNQ CPU SS block
        diagram and the firmware running on the zynq PS need to be handled.
        """

        # Generate register map according to IO settings stored in structure config and add to target sources
        with (open(self._simctrlregmap_path, 'w')) as ctrl_file:
           ctrl_file.write(ModuleRegMapSimCtrl(scfg=self.scfg).render())

        content.verilog_sources += [VerilogSource(files=self._simctrlregmap_path, name='simctrlregmap')]

        # Generate Hardware Abstraction Layer according to control signals
        # specified in the simctrl.yaml file and add to sources.

        gpio_fw = FirmwareGPIO(scfg=self.scfg)

        # Write header code
        gpio_hdr = os.path.join(self.pcfg.build_root, 'gpio_funcs.h')
        with open(gpio_hdr, 'w') as f:
            f.write(gpio_fw.hdr_text)
        content.firmware_files += [FirmwareFile(files=gpio_hdr, name='gpio_hdr')]

        # Write implementation code
        gpio_src = os.path.join(self.pcfg.build_root, 'gpio_funcs.c')
        with open(gpio_src, 'w') as f:
            f.write(gpio_fw.src_text)
        content.firmware_files += [FirmwareFile(files=gpio_src, name='gpio_src')]

        # Generate application code for UART_ZYNQ control, if no custom code is provided
        if not self.tcfg.custom_zynq_firmware:
            appcode = UartZynqFirmwareAppCode(scfg=self.scfg)
            # Write application code and add to firmware files
            appcode_src = os.path.join(self.pcfg.build_root, 'main.c')
            with open(appcode_src, 'w') as f:
                f.write(appcode.src_text)
            content.firmware_files += [FirmwareFile(files=appcode_src, name='appcode_src')]

    def add_ip_cores(self, scfg, ip_dir):
        """
        Configures and adds IP cores that are necessary for selected IP cores. VIO IP core is configured and added.
        :return rendered template for configuring a vio IP core
        """
        return []


from anasymod.sim_ctrl.ctrlinfra import ControlInfrastructure
from anasymod.structures.module_viosimctrl import ModuleVIOSimCtrl
from anasymod.sources import VerilogSource
from anasymod.structures.structure_config import StructureConfig
from anasymod.templates.vio_wiz import TemplVIO

class VIOControlInfrastructure(ControlInfrastructure):
    def __init__(self, prj_cfg, plugin_includes):
        super().__init__(prj_cfg=prj_cfg, plugin_includes=plugin_includes)

        # Initialize internal variables

    def gen_ctrlwrapper(self, str_cfg: StructureConfig, content):
        """
        Generate RTL design for base control infrastructure. This will generate the sim ctrl wrapper for VIO control.
        """

        # Generate simulation control wrapper and add to target sources
        with (open(self._simctrlwrap_path, 'w')) as ctrl_file:
           ctrl_file.write(ModuleVIOSimCtrl(scfg=str_cfg, plugin_includes=self.plugin_includes).render())

        content.verilog_sources += [VerilogSource(files=self._simctrlwrap_path, name='simctrlwrap')]

    def gen_ctrl_infrastructure(self, content):
        """
        Generate RTL design for FPGA specific control infrastructure, depending on the interface selected for communication.
        For VIO control, no additional RTL sources need to be generated.

        """
        pass

    def add_ip_cores(self, scfg, ip_dir):
        """
        Configures and adds IP cores that are necessary for selected IP cores. VIO IP core is configured and added.
        :return rendered template for configuring a vio IP core
        """
        return [TemplVIO(scfg=scfg, ip_dir=ip_dir)]


def main():
    ctrl = VIOControlInfrastructure(prj_cfg=EmuConfig(root='test', cfg_file=''))
    ctrl.gen_ctrlwrapper(str_cfg=StructureConfig(prj_cfg=EmuConfig(root='test', cfg_file='')), content='')

if __name__ == "__main__":
    main()


from anasymod.templates.templ import JinjaTempl
from anasymod.config import EmuConfig
from anasymod.generators.gen_api import SVAPI, ModuleInst
from anasymod.structures.structure_config import StructureConfig
from anasymod.sim_ctrl.datatypes import DigitalSignal


class ModuleEmuClks(JinjaTempl):
    def __init__(self, scfg: StructureConfig, pcfg: EmuConfig):
        super().__init__(trim_blocks=True, lstrip_blocks=True)

        gated_clks = []

        #####################################################
        # Create module interface
        #####################################################
        self.module_ifc = SVAPI()

        module = ModuleInst(api=self.module_ifc, name="gen_emu_clks")
        module.add_input(scfg.emu_clk_2x)
        module.add_output(scfg.emu_clk)

        for derived_clk in scfg.clk_derived:
            if derived_clk.abspath_gated_clk is not None:
                gated_clks.append(derived_clk.name)

        # add IOs for default oscillator if used
        self.use_default_oscillator = scfg.use_default_oscillator
        if scfg.use_default_oscillator:
            module.add_input(DigitalSignal(name=f'clk_val_default_osc', width=1, abspath=''))
            module.add_output(DigitalSignal(name=f'clk_default_osc', width=1, abspath=''))

        for gated_clk in gated_clks:
            module.add_input(DigitalSignal(name=f'clk_val_{gated_clk}', width=1, abspath=''))
            module.add_output(DigitalSignal(name=f'clk_{gated_clk}', width=1, abspath=''))

        module.generate_header()

        #####################################################
        # Generate other clks
        #####################################################
        self.generated_clks = SVAPI()

        if gated_clks:
            for gated_clk in gated_clks:
                self.generated_clks.gen_signal(DigitalSignal(name=f'clk_unbuf_{gated_clk}', width=1, abspath=''), default_value=0)
            self.generated_clks.writeln(f'always @(posedge emu_clk_2x) begin')
            self.generated_clks.indent()
            for gated_clk in gated_clks:
                self.generated_clks.writeln(f"if (emu_clk_unbuf == 1'b0) begin")
                self.generated_clks.indent()
                self.generated_clks.writeln(f'clk_unbuf_{gated_clk} <= clk_val_{gated_clk};')
                self.generated_clks.dedent()
                self.generated_clks.writeln(f'end else begin')
                self.generated_clks.indent()
                self.generated_clks.writeln(f"clk_unbuf_{gated_clk} <= clk_unbuf_{gated_clk};")
                self.generated_clks.dedent()
                self.generated_clks.writeln(f'end')
            self.generated_clks.dedent()
            self.generated_clks.writeln(f'end')
            self.generated_clks.writeln(f'')
            self.generated_clks.writeln(f'`ifndef SIMULATION_MODE_MSDSL')
            self.generated_clks.indent()
            for k, gated_clk in enumerate(gated_clks):
                self.generated_clks.writeln(f'BUFG buf_{k} (.I(clk_unbuf_{gated_clk}), .O(clk_{gated_clk}));')
            self.generated_clks.dedent()
            self.generated_clks.writeln(f'`else')
            self.generated_clks.indent()
            for gated_clk in gated_clks:
                self.generated_clks.writeln(f'assign clk_{gated_clk} = clk_unbuf_{gated_clk};')
            self.generated_clks.dedent()
            self.generated_clks.writeln(f'`endif')

    TEMPLATE_TEXT = '''
`timescale 1ns/1ps

`default_nettype none

{{subst.module_ifc.text}}

// generate emu_clk
logic emu_clk_unbuf = 0;

always @(posedge emu_clk_2x) begin
    emu_clk_unbuf <= ~emu_clk_unbuf;
end
`ifndef SIMULATION_MODE_MSDSL
    BUFG buf_emu_clk (.I(emu_clk_unbuf), .O(emu_clk));
`else
    assign emu_clk = emu_clk_unbuf;
`endif

{% if subst.use_default_oscillator %}
// Handle default oscillator
logic clk_unbuf_default_osc = 0;
always @(posedge emu_clk_2x) begin
    clk_unbuf_default_osc <= (~emu_clk_unbuf) & clk_val_default_osc;
end
`ifndef SIMULATION_MODE_MSDSL
    BUFG buf_default_osc (.I(clk_unbuf_default_osc), .O(clk_default_osc));
`else
    assign clk_default_osc = clk_unbuf_default_osc;
`endif
{% endif %}

// generate other clocks
{{subst.generated_clks.text}}
endmodule
 
`default_nettype wire
'''

def main():
    print(ModuleEmuClks(scfg=StructureConfig(prj_cfg=EmuConfig(root='test', cfg_file=''))).render())

if __name__ == "__main__":
    main()



class UartZynqFirmwareAppCode:
    def __init__(self, scfg):
        crtl_inputs = scfg.analog_ctrl_inputs + scfg.digital_ctrl_inputs
        ctrl_outputs = scfg.analog_ctrl_outputs + scfg.digital_ctrl_outputs

        self.getter_dict = {}
        for probe in ctrl_outputs:
            if probe.o_addr is not None:
                self.getter_dict[probe.o_addr] = probe.name

        self.setter_dict = {}
        for param in crtl_inputs:
            if param.i_addr is not None:
                self.setter_dict[param.i_addr] = param.name

        self.src_text = self.gen_src_text()

    def gen_src_text(self):
        retval = '''
#include "gpio_funcs.h"
#include <stdio.h>
#include <string.h>
#include "sleep.h"
'''

        idx = 0

        # add setter defines
        for k, v in self.setter_dict.items():
            retval += f'''
#define SET_{v} {idx}'''
            idx += 1

        # add getter defines
        for k, v in self.getter_dict.items():
            retval += f'''
#define GET_{v} {idx}'''
            idx += 1

        # add default body 1
        retval += r'''

int main() {
    // character buffering
    u32 idx = 0;
    char rcv;
    char buf [32];
    
    // command processing;
    u32 cmd;
    u32 arg1;
    u32 nargs = 0;

    if (init_GPIO() != 0) {
        xil_printf("GPIO Initialization Failed\r\n");
        return XST_FAILURE;
    }

    while (1) {
        rcv = inbyte();
        if ((rcv == ' ') || (rcv == '\t') || (rcv == '\r') || (rcv == '\n')) {
            // whitespace
            if (idx > 0) {
                buf[idx++] = '\0';
                if (nargs == 0) {
                    if (strcmp(buf, "HELLO") == 0) {
                        xil_printf("Hello World!\r\n");
                        nargs = 0;
                    } else if (strcmp(buf, "EXIT") == 0) {
                        return 0;
'''

        #add cmd interpreter for setters
        for k, v in self.setter_dict.items():
            retval += f'''
                    }} else if (strcmp(buf, "SET_{v}") == 0) {{
                        cmd = SET_{v};
                        nargs++;
'''

        # add cmd interpreter for getters
        for k, v in self.getter_dict.items():
            retval += f'''
                    }} else if (strcmp(buf, "GET_{v}") == 0) {{
                        xil_printf("%0d\\r\\n", get_{v}());
                        nargs = 0;
'''

        # add default body 2
        retval += r'''
                    } else {
                        xil_printf("ERROR: Unknown command\r\n");
                    }
                } else if (nargs == 1) {
                    sscanf(buf, "%lu", &arg1);
'''

        # add setter argument reader
        cond_str = 'if'
        for k, v in self.setter_dict.items():
            retval += f'''
                    {cond_str} (cmd == SET_{v}) {{
                        set_{v}(arg1)\n;
''' + \
r'''                      
                        xil_printf("0\r\n");
'''
            cond_str = '} else if'

        # add default body 3
        retval += '''
                    }
                    nargs = 0;
                }
            }
            idx = 0;
        } else {
            // load next character
            buf[idx++] = rcv;
        }
    }

    return 0;
}  
'''

        # return the code
        return retval


import sys, os
import time
from numbers import Number
from pathlib import Path
from .console_print import cprint_block
from anasymod.sim_ctrl.ctrlapi import CtrlApi
from anasymod.generators.gen_api import CodeGenerator
from anasymod.templates.launch_FPGA_sim import TemplLAUNCH_FPGA_SIM
from anasymod.structures.structure_config import StructureConfig
from anasymod.config import EmuConfig
from anasymod.enums import TraceUnitOperators
from anasymod.sim_ctrl.datatypes import AnalogProbe, DigitalSignal
from anasymod.wave import ConvertWaveform
from anasymod.util import expand_path
from anasymod.files import mkdir_p

SERVER_PORT = 57937

class VIOCtrlApi(CtrlApi):
    """
    Start an interactive control interface to HW target for running regression tests or design exploration/debug.
    For FPGA/Emulators, as a pre-requisit, bitstream must have been created and programmed. Additionally any eSW
    necessary in the targeted system must have already been programmed.
    """
    def __init__(self, result_path_raw, result_type_raw, result_path, scfg: StructureConfig, pcfg: EmuConfig,
                 bitfile_path, ltxfile_path, cwd=None, prompt='Vivado% ', err_strs=None, debug=False, float_type=False,
                 dt_scale=1e-15):
        super().__init__(pcfg=pcfg, scfg=scfg, cwd=cwd, prompt=prompt, debug=debug)
        # set defaults
        if err_strs is None:
            err_strs = ['ERROR', 'FATAL']

        self.result_path_raw = result_path_raw
        self.result_type_raw = result_type_raw
        self.result_path = result_path
        self.float_type = float_type

        # save settings
        self.err_strs = err_strs
        self.bitfile_path = bitfile_path
        self.ltxfile_path = ltxfile_path

        # create dictionary of analog control I/O
        # TODO: is there a better way to access this information?
        self.analog_ctrl_outputs = {elem.name: elem for elem in self.scfg.analog_ctrl_outputs}
        self.analog_ctrl_inputs = {elem.name: elem for elem in self.scfg.analog_ctrl_inputs}

        self.record_timeout = 0.0

    ### User Functions

    def sendline(self, line, timeout=float('inf')):
        """
        Send a single line in target shell specific language e.g. in tcl for tcl shell.
        :param line: Line that shall be send to/processed by shell
        :param timeout: Maximum time granted for operation to finish
        :return: Return string from Vivado TCL interpreter
        """
        if self.debug:
            cprint_block([line], title='SEND', color='magenta')

        self.proc.sendline(line)
        before = self._expect_prompt(timeout=timeout)

        # make sure that there were no errors
        for err_str in self.err_strs:
            if err_str in before:
                raise Exception(f'Found {err_str} in output from Vivado.')

        return before

    def source(self, script, timeout=float('inf')):
        """
        Source a script written language for targeted shell.
        :param script: Name/Path to script that shall be sourced
        :param timeout: Maximum time granted for operation to finish
        """
        script = Path(script).resolve()
        res = self.sendline(f'source {script.as_posix()}', timeout=timeout)

    def setup_trace_unit(self, trigger_name, trigger_operator, trigger_value, sample_decimation=None, sample_count=None):
        """
        Setup the trace unit. This involves defining the signal, that shall be used to start tracing and defining the
        comparison operator, which is used to monitor the trigger signal. Also arm the trace unit at the end.

        Note: In case the trigger value is set too high and an overflow occurs, the simulation will hang, as the trigger
        condition will never be met.

        :param trigger_name:        Probe signal to be used as trigger signal.
        :param trigger_operator:    Comparison operator that is used when monitoring the trigger signal to detect,
                                    when a trigger shall be fired. Available oerators are:
                                        eq (equal),
                                        neq (not equal),
                                        gt (greater than),
                                        gteq (greater or equal than),
                                        lt (lesser than),
                                        lteq (lesser or equal than)
        :param trigger_value:       Value, the probe signal is compared to.
        :param sample_decimation:   Number of samples to be skipped during recording. By default, every sample will
                                    stored in the results file, adding a sample_decimation will allow to only record
                                    every x sample.
        :param sample_count:        Number of samples to be recorded. This number shall not exceed that maximum ILA
                                    depth defined during project setup AND needs to be a multiple of 2.
        """

        # Compute timeout, that needs to be set for ila recording. Due to a bug in the ILA core, recording does not
        # return, unless a timeout is specified, for recordings exceeding 0.26 s
        self.record_timeout = 0.0

        trigger_obj = None

        if sample_count:
            # This solution I got from https://stackoverflow.com/questions/57025836/check-if-a-given-number-is-power-of-two-in-python
            # and is checking if sample_count is a power of two using bit manipulations
            if (sample_count & (sample_count-1) == 0) and sample_count != 0:
                depth = sample_count
            else:
                raise Exception(f'ERROR: Sample count needs to be a power of 2, but is set to: {sample_count}')
        else:
            depth = self.pcfg.ila_depth

        # Check, if provided trigger_signal is a valid signal connected to the ILA
        for probe in [self.scfg.time_probe] + self.scfg.analog_probes + self.scfg.digital_probes:
            if trigger_name == probe.name:
                trigger_obj = probe
            elif trigger_name == 'time':
                trigger_obj = self.scfg.time_probe
            else:
                raise Exception(f'ERROR: Provided trigger_signal:{trigger_name} is not a valid probe in the existing project!')

        # Check if operator set is legal
        if not trigger_operator in TraceUnitOperators.__dict__.values():
            raise Exception(f'ERROR: Provided operator:{trigger_operator} is not legal! Supported operators are:{TraceUnitOperators.__dict__.values()}')

        # Convert compare value depending on signal type
        if isinstance(trigger_obj, AnalogProbe):
            if not isinstance(trigger_value, float):
                raise Exception(
                    f'ERROR: trigger_value type for an analog trigger signal shall ne float, it is instead:{type(trigger_value)}')
            value_int = int(round(trigger_value * (2 ** (int(-trigger_obj.exponent)))))
            trigger_value_as_bin = f"{int(trigger_obj.width)}'u{value_int}"
        elif isinstance(trigger_obj, DigitalSignal):
            # For the time signal, conversion from float to int is necessary
            if trigger_name in ['time', self.scfg.time_probe.name]:
                # Convert trigger value to integer considering dt_scale
                trigger_value_as_int = int(float(trigger_value) / float(self.pcfg.cfg.dt_scale))
                # Represent as binary and expand to time_width
                trigger_value_as_bin = f"{trigger_obj.width}'b{bin(trigger_value_as_int).replace('b', '').zfill(self.pcfg.cfg.time_width)}"
                # Extend record_timeout according to time trigger
                self.record_timeout += float(trigger_value)
            elif isinstance(trigger_value, int):
                trigger_value_as_bin = f"{trigger_obj.width}'u{trigger_value}"
            else:
                p = set(trigger_value)
                s = {'0', '1'}

                if s == p or p == {'0'} or p == {'1'}:
                    trigger_value_as_bin = f"{trigger_obj.width}'b{trigger_value}"
                else:
                    raise Exception(f'ERROR: Provided trigger value:{trigger_value} is not valid for a digital trigger '
                                    f'signal, either provide and interger or binary string!')
        else:
            raise Exception(f'ERROR: No valid signal type for provided trigger signal:{trigger_name} Type:{type(trigger_obj)}')

        self.sendline(f'set_property CONTROL.CAPTURE_MODE BASIC $ila_0_i')
        self.sendline(f'set_property CONTROL.TRIGGER_POSITION 0 $ila_0_i')
        self.sendline(f"set_property TRIGGER_COMPARE_VALUE {trigger_operator}{trigger_value_as_bin} [get_hw_probes trace_port_gen_i/{trigger_obj.name} -of_objects $ila_0_i]")

        # Data depth is set to 2, as 1 is not supported by Vivado's ILA Core
        self.sendline(f'set_property CONTROL.DATA_DEPTH {depth} $ila_0_i')

        # Window count is set to half of the selected ILA depth
        #self.sendline(f'set_property CONTROL.WINDOW_COUNT {int(window_count)} $ila_0_i')
        self.sendline(f'set_property CONTROL.WINDOW_COUNT 1 $ila_0_i')
        self.sendline(f"set_property CAPTURE_COMPARE_VALUE eq1'b1 [get_hw_probes trace_port_gen_i/emu_dec_cmp -of_objects $ila_0_i]")

        # Set decimation threshold signal to value defined in sample_decimation if set
        if sample_decimation:
            if isinstance(sample_decimation, int): # Check if sample_decimation is an integer
                self.set_param(name=self.scfg.dec_thr_ctrl.name, value=sample_decimation, timeout=30)
            else:
                raise Exception(f'Provided sample_decimation value is of wrong type, expecting integer and got:{type(sample_decimation)}')
        else:
            sample_decimation = 0

        self.arm_trace_unit()

        # conclude calculation of record_timeout by considering time needed to collect samples and representing
        # over minutes

        self.record_timeout += float(self.pcfg.cfg.dt) * depth * sample_decimation
        self.record_timeout = self.record_timeout / 60

    def arm_trace_unit(self):
        """
        Arm the trace unit, this will delete the buffer and arm the trigger.
        """
        self.sendline('run_hw_ila $ila_0_i')

    def wait_on_and_dump_trace(self, result_file=None, emu_time_scaled=True):
        """
        Wait until the trace unit stopped recording data. Transmit this data to the host PC, store by default to the raw
        result file path, or a custom path provided by the user, and convert analog values from fixed-point to float.
        Finally store it to a .vcd file in the default location.

        :param result_file: Optionally, it is possible to provide a custom result file path.
        """

        if result_file is not None:
            # Expand provided path, paths relative to project root are also supported
            result_path = expand_path(result_file, rel_path_reference=self.pcfg.root)

            # Create raw result path by adding _raw to the filename
            result_path_raw = os.path.join(os.path.dirname(result_path),
                                           os.path.basename(os.path.splitext(result_path)[0]) + '_raw' +
                                           os.path.splitext(result_path)[1])
            print(f'Simulation results will be stored in:{result_path}')
        else:
            result_path = self.result_path
            result_path_raw = self.result_path_raw

        if not result_path:
            raise Exception(f'ERROR: provided result_file:{result_file} is not valid!')

        # wait until trace buffer is full
        self.sendline(f'wait_on_hw_ila -timeout {self.record_timeout} $ila_0_i')

        # transmit and dump trace buffer data to a CSV file
        self.sendline('current_hw_ila_data [upload_hw_ila_data $ila_0_i]')

        if not os.path.isdir(os.path.dirname(result_path_raw)):
            mkdir_p(os.path.dirname(result_path_raw))

        self.sendline(f'write_hw_ila_data -csv_file -force {{{result_path_raw}}} [current_hw_ila_data]')

        # Convert to .vcd and from fixed-point to float
        ConvertWaveform(result_path_raw=result_path_raw,
                        result_type_raw=self.result_type_raw,
                        result_path=result_path,
                        str_cfg=self.scfg,
                        float_type=self.float_type,
                        dt_scale=self.pcfg.cfg.dt_scale,
                        emu_time_scaled=emu_time_scaled)

    def refresh_param(self, name, timeout=30):
        """
        Refresh selected control parameter.
        :param name: Name of control parameter
        :param timeout: Maximum time granted for operation to finish
        """
        self.sendline(f'refresh_hw_vio ${name}', timeout=timeout)

    def get_param(self, name, timeout=30):
        """
        Read value of a control parameter in design.
        :param name: Name of control parameter to be read
        :param timeout: Maximum time granted for operation to finish
        """
        # get output result as a string
        value = self.sendline(f'get_property INPUT_VALUE ${name}', timeout=timeout)
        value = value.splitlines()[-1] # get last line
        value = value.strip() # strip off whitespace

        # convert value to floating-point if needed
        if name in self.analog_ctrl_outputs:
            value = self.analog_ctrl_outputs[name].fixed_to_float(int(value))

        # return value
        return value

    def set_param(self, name, value, timeout=30):
        """
        Set value of a control parameter in design.
        :param name: Name of control parameter to be set
        :param value: Value of control parameter sto be set
        :param timeout: Maximum time granted for operation to finish
        """
        # convert value to fixed-point if needed
        if name in self.analog_ctrl_inputs:
            value = self.analog_ctrl_inputs[name].float_to_fixed(value)

        # send command
        self.sendline(f'set_property OUTPUT_VALUE {value} ${name}', timeout=timeout)
        self.sendline(f'commit_hw_vio ${name}')

    def set_var(self, name, value):
        """
        Define a variable in target shell environment.
        :param name: Name of variable that shall be set
        :param value: Value of variable that shall be set
        """
        self.sendline(f'set {name} {self._tcl_val(value)}')

    def set_reset(self, value, timeout=30):
        """
        Control the 'emu_rst' signal, in order to put the system running on the FPGA into or out of reset state.
        :param value: Value of reset signal, 1 will set it to reset and 0 will release reset.
        :param timeout: Maximum time granted for operation to finish
        """
        self.set_param(name=self.scfg.reset_ctrl.name, value=value, timeout=timeout)

    def get_emu_time_int(self, timeout=30):
        """
        Get current time of the FPGA simulation as an unscaled integer value.
        :param timeout: Maximum time granted for operation to finish
        """
        self.refresh_param('vio_0_i')
        emu_time_vio = self.get_param(name=self.scfg.emu_time_vio.name, timeout=timeout)
        return int(emu_time_vio)

    ### Utility Functions

    @classmethod
    def _tcl_val(cls, value):
        if isinstance(value, (list, tuple)):
            return '[list ' + ' '.join(cls._tcl_val(elem) for elem in value) + ']'
        elif isinstance(value, str):
            return '"' + value + '"'
        elif isinstance(value, Path):
            return cls._tcl_val(str(value))
        elif isinstance(value, Number):
            return str(value)
        else:
            raise Exception(f"Don't know how to convert to a TCL literal: {value}.")

    def _initialize(self):
        """
        Initialize the control interface, this is usually done after the bitstream was programmed successfully on the FPGA.
        :return:
        """

        super()._initialize_vivado_tcl()

    def _setup_ctrl(self, server_addr):
        """
        Prepare instrumentation on the FPGA to allow interactive control.
        :param server_addr: Address of remote hardware server
        :return:
        """
        launch_script = os.path.join(os.path.dirname(os.path.dirname(self.result_path_raw)), r"launch_FPGA.tcl")
        codegen = CodeGenerator()
        codegen.use_templ(TemplLAUNCH_FPGA_SIM(pcfg=self.pcfg, scfg=self.scfg, bitfile_path=self.bitfile_path,
                                               ltxfile_path=self.ltxfile_path, server_addr=server_addr))
        codegen.write_to_file(launch_script)
        self.source(script=launch_script)

    def __del__(self):
        try:
            print('Sending "exit" to Vivado TCL interpreter.')
            self.proc.sendline('exit')
        except:
            print('Could not send "exit" to Vivado TCL interpreter.')

def get_vivado_tcl_client():
    import xmlrpc.client
    return xmlrpc.client.ServerProxy(f'http://localhost:{SERVER_PORT}')

def main():
    # modified from https://docs.python.org/3.7/library/xmlrpc.server.html?highlight=xmlrpc
    print(f'Launching Vivado TCL server on port {SERVER_PORT}.')

    from xmlrpc.server import SimpleXMLRPCServer
    from xmlrpc.server import SimpleXMLRPCRequestHandler

    # Restrict to a particular path.
    class RequestHandler(SimpleXMLRPCRequestHandler):
        rpc_paths = ('/RPC2',)

    # Instantiate TCL evaluator
    tcl = VIOCtrlApi()

    # Create server
    with SimpleXMLRPCServer(('localhost', SERVER_PORT),
                            requestHandler=RequestHandler,
                            allow_none=True) as server:
        server.register_introspection_functions()

        # list of functions available to the client
        server.register_function(tcl.sendline)
        server.register_function(tcl.source)
        server.register_function(tcl.refresh_param)
        server.register_function(tcl.set_param)
        server.register_function(tcl.get_param)

        # program not progress past this point unless
        # Ctrl-C or similar is pressed.
        server.serve_forever()

if __name__ == '__main__':
    main()



import os

from anasymod.viewer.viewer import Viewer
from anasymod.util import call

class GtkWaveViewer(Viewer):
    def view(self, result_file=None):
        vcd_path = result_file if result_file is not None else self.target.cfg.vcd_path

        # build command
        if os.path.isfile(vcd_path):
            cmd = [self.cfg.gtkwave_config.gtkwave, vcd_path]
            cmd[1:1] = self.cfg.gtkwave_config.lsf_opts.split()
            # add waveform file if it exists
            if self.cfg.gtkwave_config.gtkw_config is not None:
                if os.path.isfile(self.cfg.gtkwave_config.gtkw_config):
                    cmd.append(self.cfg.gtkwave_config.gtkw_config)

            # run command
            call(cmd)
        else:
            raise Exception(f'ERROR: Result file: {vcd_path} does not exist; cannot open waveforms!')


from anasymod.templates.templ import JinjaTempl
from anasymod.generators.gen_api import SVAPI, ModuleInst
from anasymod.sim_ctrl.datatypes import DigitalSignal
from anasymod.structures.structure_config import StructureConfig
from anasymod.templates.zynq_gpio import TemplZynqGPIO

class ModuleUARTSimCtrl(JinjaTempl):
    def __init__(self, scfg: StructureConfig):
        super().__init__(trim_blocks=True, lstrip_blocks=True)

        ctrl_inputs = scfg.analog_ctrl_inputs + scfg.digital_ctrl_inputs
        ctrl_outputs = scfg.analog_ctrl_outputs + scfg.digital_ctrl_outputs

        #####################################################
        # Define module ios
        #####################################################
        self.module_ifc = SVAPI()
        self.zynq_gpio = TemplZynqGPIO()

        module = ModuleInst(api=self.module_ifc, name="sim_ctrl_gen")
        module.add_inputs(ctrl_outputs)
        module.add_outputs(ctrl_inputs)
        module.add_input(scfg.emu_clk)
        if not scfg.is_ultrascale:
            module.add_inouts(TemplZynqGPIO.EXT_IOS)
        module.generate_header()

        ctrl_inputs = scfg.analog_ctrl_inputs + scfg.digital_ctrl_inputs
        ctrl_outputs = scfg.analog_ctrl_outputs + scfg.digital_ctrl_outputs

        #####################################################
        # PC sim control section
        #####################################################

        ## Custom control IOs for pc sim control module
        self.pc_sim_crtl_ifc = SVAPI()

        # Only generate instantiation if there is any custom ctrl signal available
        ctrl_signals = ctrl_outputs + ctrl_inputs
        custom_ctrl_signals = []
        for ctrl_signal in ctrl_signals:
            if not ctrl_signal.name in scfg.special_ctrl_ios:
                custom_ctrl_signals.append(ctrl_signal)

        if len(custom_ctrl_signals) != 0:
            sim_ctrl_module = ModuleInst(api=self.pc_sim_crtl_ifc, name="sim_ctrl")
            for ctrl_output in ctrl_outputs:
                if ctrl_output.name not in scfg.special_ctrl_ios:
                    sim_ctrl_module.add_input(ctrl_output, connection=ctrl_output)
            for ctrl_input in ctrl_inputs:
                if ctrl_input.name not in scfg.special_ctrl_ios:
                    sim_ctrl_module.add_output(ctrl_input, connection=ctrl_input)

            sim_ctrl_module.generate_instantiation()

        # set number of clk cycles for initial reset
        self.rst_clkcycles = scfg.cfg.rst_clkcycles

        #####################################################
        # FPGA sim control section
        #####################################################

        # instantiation of register map module

        self.reg_map_inst = SVAPI()

        reg_map = ModuleInst(api=self.reg_map_inst, name="reg_map")

        reg_map.add_input(io_obj=DigitalSignal(name='clk', width=1, abspath=None),
                          connection=DigitalSignal(name='emu_clk', width=1, abspath=None))

        reg_map.add_input(io_obj=self.zynq_gpio.o_ctrl, connection=self.zynq_gpio.o_ctrl)
        reg_map.add_output(io_obj=self.zynq_gpio.o_data, connection=self.zynq_gpio.o_data)
        reg_map.add_input(io_obj=self.zynq_gpio.i_ctrl, connection=self.zynq_gpio.i_ctrl)
        reg_map.add_input(io_obj=self.zynq_gpio.i_data, connection=self.zynq_gpio.i_data)

        reg_map.add_outputs(io_objs=ctrl_inputs, connections=ctrl_inputs)
        reg_map.add_inputs(io_objs=ctrl_outputs, connections=ctrl_outputs)

        reg_map.generate_instantiation()

        # instantiation of bd

        self.bd_inst = SVAPI()

        bd = ModuleInst(api=self.bd_inst, name=self.zynq_gpio.design_name)

        bd.add_output(io_obj=self.zynq_gpio.o_ctrl, connection=self.zynq_gpio.o_ctrl)
        bd.add_input(io_obj=self.zynq_gpio.o_data, connection=self.zynq_gpio.o_data)
        bd.add_output(io_obj=self.zynq_gpio.i_ctrl, connection=self.zynq_gpio.i_ctrl)
        bd.add_output(io_obj=self.zynq_gpio.i_data, connection=self.zynq_gpio.i_data)

        if not scfg.is_ultrascale:
            bd.add_inouts(io_objs=TemplZynqGPIO.EXT_IOS, connections=TemplZynqGPIO.EXT_IOS)

        bd.generate_instantiation()

    TEMPLATE_TEXT = '''\
`timescale 1ns/1ps

`include "svreal.sv"

`default_nettype none
{{subst.module_ifc.text}}

`ifdef SIMULATION_MODE_MSDSL
    // reset sequence
    logic emu_rst_state = 1'b1;    
    assign emu_rst = emu_rst_state;
    initial begin
        @(posedge emu_clk);
        #((0.1/(`EMU_CLK_FREQ))*1s);
        emu_rst_state = 1'b0;
    end

    // decimation threshold
    `ifndef DEC_THR_VAL_MSDSL
        `define DEC_THR_VAL_MSDSL 0
    `endif // `ifdef DEC_THR_VAL_MSDSL
    assign emu_dec_thr = `DEC_THR_VAL_MSDSL;

    // stall / run / sleep controls
    logic [((`TIME_WIDTH)-1):0] emu_ctrl_data_state;
    logic [3:0] emu_ctrl_mode_state;
    assign emu_ctrl_data = emu_ctrl_data_state;
    assign emu_ctrl_mode = emu_ctrl_mode_state;

    // module for custom vio handling
    // NOTE: sim_ctrl module must be written and added to the project manually!!!
{{subst.pc_sim_crtl_ifc.text}}
`else
    // Signal declaration
    logic [31:0] o_ctrl;
    logic [31:0] o_data;
    logic [31:0] i_ctrl;
    logic [31:0] i_data;

    // Instantiation of register map
    {{subst.reg_map_inst.text}}

    // Instantiation of processing system
    {{subst.bd_inst.text}}
`endif // `ifdef SIMULATION_MODE_MSDSL

endmodule

`default_nettype wire
'''


from anasymod.templates.templ import JinjaTempl
from anasymod.config import EmuConfig
from anasymod.generators.gen_api import SVAPI, ModuleInst
from anasymod.structures.structure_config import StructureConfig
from anasymod.sim_ctrl.datatypes import DigitalSignal


class ModuleTimeManager(JinjaTempl):
    def __init__(self, scfg: StructureConfig, pcfg: EmuConfig, plugin_includes: list):
        super().__init__(trim_blocks=True, lstrip_blocks=True)

        self.num_dt_reqs = scfg.num_dt_reqs
        self.dt_value = pcfg.cfg.dt

        #####################################################
        # Add plugin specific includes
        #####################################################

        self.plugin_includes = SVAPI()
        for plugin in plugin_includes:
            for include_statement in plugin.include_statements:
                self.plugin_includes.writeln(f'{include_statement}')

        #####################################################
        # Create module interface
        #####################################################
        self.module_ifc = SVAPI()

        module = ModuleInst(api=self.module_ifc, name="gen_time_manager")
        module.add_input(scfg.emu_clk)
        module.add_input(scfg.reset_ctrl)
        module.add_output(scfg.time_probe)

        module.add_output(DigitalSignal(name=f'emu_dt', abspath='', width=pcfg.cfg.dt_width, signed=False))

        # add inputs for external timestep requests
        dt_reqs = []
        for derived_clk in scfg.clk_derived:
            if derived_clk.abspath_dt_req is not None:
                dt_req = DigitalSignal(
                    name=f'dt_req_{derived_clk.name}',
                    abspath='',
                    width=pcfg.cfg.dt_width,
                    signed=False
                )
                module.add_input(dt_req)
                dt_reqs.append(dt_req)

        # add input for anasymod control dt request signal
        dt_req = DigitalSignal(name=f'dt_req_stall', abspath='', width=pcfg.cfg.dt_width, signed=False)
        module.add_input(dt_req)
        dt_reqs.append(dt_req)

        # add input for dt request signal, in case a default oscillator is used
        if scfg.use_default_oscillator:
            dt_req = DigitalSignal(name=f'dt_req_default_osc', abspath='', width=pcfg.cfg.dt_width, signed=False)
            module.add_input(dt_req)
            dt_reqs.append(dt_req)

        module.generate_header()

        # generate a bit of code to take the minimum of the timestep requests
        self.codegen = SVAPI()
        self.codegen.indent()

        # take minimum of the timestep requests
        if len(dt_reqs) == 0:
            # Convert dt value to integer considering dt_scale
            dt_as_int = int(float(pcfg.cfg.dt) / float(pcfg.cfg.dt_scale))

            # Represent as binary and expand to dt_width
            dt_as_bin = bin(dt_as_int).replace('b', '').zfill(pcfg.cfg.dt_width)

            # assign to the emulator timestep output
            self.codegen.writeln(f"assign emu_dt = {pcfg.cfg.dt_width}'b{dt_as_bin};")
        else:
            prev_min = None
            for k, curr_sig in enumerate(dt_reqs):
                if k == 0:
                    prev_min = curr_sig.name
                else:
                    # create a signal to hold temporary min result
                    curr_min = f'__dt_req_min_{k - 1}'
                    self.codegen.writeln(f'(* dont_touch = "true" *) logic [((`DT_WIDTH)-1):0] {curr_min};')
                    # take the minimum of the previous minimum and the current signal
                    curr_min_val = self.vlog_min(curr_sig.name, prev_min)
                    self.codegen.writeln(f'assign {curr_min} = {curr_min_val};')
                    # mark the current minimum as the previous minimum for the next
                    # iteration of the loop
                    prev_min = curr_min
            # assign to the emulator timestep output
            self.codegen.writeln(f'assign emu_dt = {prev_min};')

    @staticmethod
    def vlog_min(a, b):
        return f'((({a}) < ({b})) ? ({a}) : ({b}))'

    TEMPLATE_TEXT = '''\
`timescale 1ns/1ps

{{subst.plugin_includes.text}}

`default_nettype none

{{subst.module_ifc.text}}
{{subst.codegen.text}}
    // assign internal state variable to output
    logic [((`TIME_WIDTH)-1):0] emu_time_state;
    assign emu_time = emu_time_state;

    // update emulation time on each clock cycle
    always @(posedge emu_clk) begin
        if (emu_rst==1'b1) begin
            emu_time_state <= 0;
        end else begin
            emu_time_state <= emu_time_state + emu_dt;
        end
    end
endmodule

`default_nettype wire
'''


def main():
    print(ModuleTimeManager(scfg=StructureConfig(prj_cfg=EmuConfig(root='test', cfg_file=''))).render())


if __name__ == "__main__":
    main()


from anasymod.enums import FPGASimCtrl

class FPGA_Board():
    board_part = None
    dbg_hub_clk_freq = 100e6
    uart_vid = None
    uart_pid = None
    uart_suffix = None
    is_ultrascale = False
    no_rev_check = False

class PYNQ_Z1(FPGA_Board):
    """
    Container to store PYNQ_Z1 FPGA board specific properties.
    """
    clk_pin = ['H16']
    clk_io = 'LVCMOS33'
    clk_freq = 125e6
    board_part = 'www.digilentinc.com:pynq-z1:part0:1.0'
    full_part_name = 'xc7z020clg400-1'
    short_part_name = 'xc7z020'
    bram = 4.9e6
    dbg_hub_clk_freq = 100e6
    fpga_sim_ctrl = [FPGASimCtrl.UART_ZYNQ, FPGASimCtrl.VIVADO_VIO]
    uart_zynq_vid = [1027]
    uart_zynq_pid = [24592]
    uart_suffix = '.1'  # needed since this board has two com ports under the same VID/PID

class ARTY_A7(FPGA_Board):
    """
    Container to store ARTY_A7 FPGA board specific properties.
    """
    clk_pin = ['E3']
    clk_io = 'LVCMOS33'
    clk_freq = 100e6
    full_part_name = 'xc7a35ticsg324-1L'
    short_part_name = 'xc7a35'
    bram = 1.8e6
    dbg_hub_clk_freq = 100e6
    fpga_sim_ctrl = [FPGASimCtrl.VIVADO_VIO]

class ARTY_200T_CUSTOM_LIDAR(FPGA_Board):
    """
    Container to store ARTY_200T_CUSTOM_LIDAR FPGA board specific properties; this is a custom board specifically for
    the Lidar Project.
    """
    clk_pin = ['H4', 'G4']
    clk_io = 'DIFF_SSTL15'
    clk_freq = 50e6
    full_part_name = 'xc7a200tfbg484-2L'
    short_part_name = 'xc7a200'
    bram = 13e6
    dbg_hub_clk_freq = 100e6
    fpga_sim_ctrl = [FPGASimCtrl.VIVADO_VIO]

class TE0720(FPGA_Board):
    """
    Container to store PYNQ_Z1 FPGA board specific properties.
    """
    clk_pin = ['F7']
    clk_io = 'LVCMOS33'
    clk_freq = 33.333333e6
    full_part_name = 'xc7z020clg484-1'
    short_part_name = 'xc7z020'
    bram = 4.9e6
    fpga_sim_ctrl = [FPGASimCtrl.UART_ZYNQ, FPGASimCtrl.VIVADO_VIO]
    uart_zynq_vid = [1027] # might need adjustment
    uart_zynq_pid = [24592] # might need adjustment

class VC707(FPGA_Board):
    """
    Container to store VC707 FPGA board specific properties.
    """
    clk_pin = ['E19', 'E18']
    clk_io = 'LVDS'
    clk_freq = 200e6
    full_part_name = 'XC7VX485T-2FFG1761C'
    short_part_name = 'XC7VX485T'
    bram = 37e6
    dbg_hub_clk_freq = 100e6
    fpga_sim_ctrl = [FPGASimCtrl.VIVADO_VIO]

class ZC702(FPGA_Board):
    """
    Container to store ZC702 FPGA board specific properties.
    """
    clk_pin = ['D18', 'C19']
    clk_io = 'LVDS_25'
    clk_freq = 200e6
    board_part = 'xilinx.com:zc702:part0:1.4'
    full_part_name = 'xc7z020clg484-1'
    short_part_name = 'xc7z020'
    bram = 4.9e6
    dbg_hub_clk_freq = 100e6
    fpga_sim_ctrl = [FPGASimCtrl.VIVADO_VIO]
    uart_zynq_vid = [4292]
    uart_zynq_pid = [60000]

class ULTRA96(FPGA_Board):
    """
    Container to store ULTRA96 FPGA board specific properties.
    """
    clk_pin = ['L19', 'L20']
    clk_io = 'LVDS'
    clk_freq = 26e6
    full_part_name = 'xczu3eg-sbva484-???'
    short_part_name = 'xczu3eg'
    bram = 7.6e6
    fpga_sim_ctrl = [FPGASimCtrl.UART_ZYNQ, FPGASimCtrl.VIVADO_VIO]
    uart_zynq_vid = [1027] # might need adjustment
    uart_zynq_pid = [24592] # might need adjustment

class ZC706(FPGA_Board):
    """
    Container to store ZC706 FPGA board specific properties.
    """
    clk_pin = ['H9', 'G9']
    clk_io = 'LVDS'
    clk_freq = 200e6
    board_part = 'xilinx.com:zc706:part0:1.4'
    full_part_name = 'xc7z045ffg900-2'
    short_part_name = 'xc7z045'
    bram = 19.2e6
    dbg_hub_clk_freq = 100e6
    fpga_sim_ctrl = [FPGASimCtrl.UART_ZYNQ, FPGASimCtrl.VIVADO_VIO]
    uart_zynq_vid = [4292]
    uart_zynq_pid = [60000]

class ZCU102(FPGA_Board):
    """
    Container to store ZCU102 FPGA board specific properties.
    """
    clk_pin = ['AL8', 'AL7']
    clk_io = 'DIFF_SSTL12'
    clk_freq = 300e6
    board_part = 'xilinx.com:zcu102:part0:3.2'
    full_part_name = 'xczu9eg-ffvb1156-2-e'
    short_part_name = 'xczu9'
    bram = 32.1e6
    fpga_sim_ctrl = [FPGASimCtrl.UART_ZYNQ, FPGASimCtrl.VIVADO_VIO]
    uart_zynq_vid = [4292]
    uart_zynq_pid = [60017]
    uart_suffix = '.0'  # needed since this board has four com ports under the same VID/PID
    is_ultrascale = True
    no_rev_check = True  # needed in case the FPGA is an engineering sample

class ZCU106(FPGA_Board):
    """
    Container to store ZCU106 FPGA board specific properties.
    """
    clk_pin = ['H9', 'G9']
    clk_io = 'LVDS'
    clk_freq = 125e6
    board_part = 'xilinx.com:zcu106:part0:2.5'
    full_part_name = 'xczu7ev-ffvc1156-2-e'
    short_part_name = 'xczu7ev'
    bram = 11.0e6
    fpga_sim_ctrl = [FPGASimCtrl.UART_ZYNQ, FPGASimCtrl.VIVADO_VIO]
    uart_zynq_vid = [4292]
    uart_zynq_pid = [60017]
    uart_suffix = '.0'  # needed since this board has four com ports under the same VID/PID
    is_ultrascale = True

class ZCU111(FPGA_Board):
    """
    Container to store ZCU111 FPGA board specific properties.
    """
    clk_pin = ['AL17', 'AM17']  # from Table 3-17 in the user guide
    clk_io = 'LVDS'
    clk_freq = 125e6
    board_part = 'xilinx.com:zcu111:part0:1.2'
    full_part_name = 'xczu28dr-ffvg1517-2-e'
    short_part_name = 'xczu28'
    bram = 38e6
    fpga_sim_ctrl = [FPGASimCtrl.UART_ZYNQ, FPGASimCtrl.VIVADO_VIO]
    uart_zynq_vid = [1027]  # FT4232HL from Table 3-8 in the user guide
    uart_zynq_pid = [24593]  # FT4232HL from Table 3-8 in the user guide
    uart_suffix = '.1'  # needed since this board has four com ports under the same VID/PID
    is_ultrascale = True
    no_rev_check = True  # needed in case the FPGA is an engineering sample



class TemplXSCTBuild:
    def __init__(self, sdk_path, version_year, version_number,
                 proc_name='ps7_cortexa9_0', os_name='standalone',
                 hw_name='hw', hw_path=None, sw_name='sw',
                 template_name='Empty Application', create=True,
                 build=True):

        # save version information
        self.version_year = version_year
        self.version_number = version_number

        # initialize text
        self.text = ''

        # apply commands

        self.setws(sdk_path=sdk_path)

        if create:
            self.app_create(
                hw_name=hw_name,
                hw_path=hw_path,
                sw_name=sw_name,
                proc_name=proc_name,
                template_name=template_name,
                os_name=os_name
            )

        if build:
            self.app_build(
                sw_name=sw_name
            )

    def line(self, s='', nl='\n'):
        self.text += f'{s}{nl}'

    def comment(self, s):
        self.line(f'# {s}')

    def setws(self, sdk_path):
        self.comment('set the work directory')
        self.line(f'setws "{sdk_path.as_posix()}"')
        self.line()

    def app_create(self, hw_name, hw_path, sw_name, proc_name,
                   template_name, os_name):
        if self.version_year < 2020:
            hw_path_hdf = hw_path.with_suffix('.hdf')
            self.comment('create the hardware configuration')
            self.line(f'file copy -force "{hw_path.as_posix()}" "{hw_path_hdf.as_posix()}"')
            self.line(f'createhw -name {hw_name} -hwspec "{hw_path_hdf.as_posix()}"')
            #self.line(f'createhw -name {hw_name} -hwspec "C:/Inicio_dev/fpga_flow/anasymod/tests/firmware/build/fpga/prj/prj.sdk/top.hdf"')
            self.line()
            self.comment('create the software configuration')
            self.line(f'createapp -name {sw_name} -hwproject {hw_name} '
                      f'-proc {proc_name} -app "{template_name}"')
            self.line()
        else:
            self.comment('create the app configuration')
            self.line(f'app create -name {sw_name} -hw "{hw_path.as_posix()}" -os {os_name}'
                      f' -proc {proc_name} -template "{template_name}"')
            self.line()

    def app_build(self, sw_name):
        self.comment('build application')
        if self.version_year < 2020:
            self.line('projects -build')
        else:
            self.line(f'app build -name {sw_name}')
        self.line()


from anasymod.base_config import BaseConfig
from anasymod.enums import ConfigSections
from anasymod.sim_ctrl.datatypes import DigitalCtrlInput, DigitalCtrlOutput, AnalogCtrlInput, AnalogCtrlOutput
from anasymod.structures.structure_config import StructureConfig

import serial, os

class ControlInfrastructure():
    def __init__(self, prj_cfg, plugin_includes):

        self.pcfg = prj_cfg
        self._simctrlwrap_path = os.path.join(self.pcfg.build_root, 'gen_ctrlwrap.sv')
        self.plugin_includes = plugin_includes

    def gen_ctrlwrapper(self, str_cfg: StructureConfig, content):
        """
        Generate RTL design for base control infrastructure, depending on the interface selected for communication.
        """
        raise NotImplementedError("This function cannot be called from the base control class itself and is overloaded "
                                  "in the inheriting classes.")

    def gen_ctrl_infrastructure(self, content):
        """
        Generate RTL design for FPGA specific control infrastructure, depending on the interface selected for communication.
        """
        raise NotImplementedError("This function cannot be called from the base control class itself and is overloaded "
                                  "in the inheriting classes.")

    def add_ip_cores(self, scfg, ip_dir):
        """
        Configures and adds IP cores that are necessary for selected IP cores.
        """
        raise NotImplementedError("This function cannot be called from the base control class itself and is overloaded "
                                  "in the inheriting classes.")


import os

from anasymod.sources import Sources, VerilogHeader, VerilogSource, VHDLSource
from anasymod.defines import Define
from anasymod.enums import ConfigSections
from anasymod.base_config import BaseConfig

class Plugin():
    def __init__(self, cfg_file, prj_root, build_root, name):
        self.args = None

        self._cfg_file = cfg_file
        self._prj_root = prj_root
        self._build_root = build_root
        self._srccfg_path = os.path.join(self._prj_root, r"source.yaml")
        self._name = name
        self._defines = []
        self.generator_sources = []
        self._verilog_sources = []
        """:type : List[VerilogSource]"""

        self._verilog_headers = []
        """:type : List[VerilogHeader]"""

        self._vhdl_sources = []
        """:type : List[VHDLSource]"""

        # List of includes that need to be added to source files generated via ANASYMOD
        self.cfg = Config(cfg_file=self._cfg_file)

        self.include_statements = []


    #### User Functions ####

    def models(self):
        """
        Runs source code generator.
        """
        raise NotImplementedError()

    def set_option(self, name, value=None):
        func = getattr(self, name)
        if callable(func):
            if value is True or None:
                func()
            elif not value:
                pass
            else:
                func(value)
        else:
            raise Exception(f'ERROR: Provided option:{name} is not supported for this generator.')

    #### Utility Functions ####

    def _add_source(self, source: Sources):
        if isinstance(source, VerilogSource):
            self._verilog_sources.append(source)
        if isinstance(source, VerilogHeader):
            self._verilog_headers.append(source)
        if isinstance(source, VHDLSource):
            self._vhdl_sources.append(source)

    def _add_define(self, define: Define):
        self._defines.append(define)

    def _dump_defines(self):
        return self._defines

    def _dump_verilog_sources(self):
        return self._verilog_sources

    def _dump_verilog_headers(self):
        return self._verilog_headers

    def _dump_vhdl_sources(self):
        return self._vhdl_sources

    def _setup_sources(self):
        """
        Add Source objects that are specific to MSDSL
        """
        raise NotImplementedError()

    def _setup_defines(self):
        """
        Add Define objects that are specific to MSDSL
        """
        raise NotImplementedError()

    def _parse_args(self):
        """
        Read command line arguments. This supports convenient usage from command shell e.g.:
        python analysis.py -i filter --models --sim --view
        """
        pass

    def _return_args(self):
        return self.args

    def _set_generator_sources(self, generator_sources: list):
        """
        Set, which functional models shall be generated via the msdsl plugin. This works by setting the class instance
        attribute self.generator_sources.

        :param generator_sources: List of source objects specific to the generator
        """

        if any(isinstance(i, Sources) for i in generator_sources): # Make sure all elements in the list are of type Sources
            [i.expand_paths() for i in generator_sources]
            self.generator_sources = generator_sources
        else:
            raise Exception(f'ERROR: Format for argument generator_sources is incorrect, expected list, got:{type(generator_sources)}')

class Config(BaseConfig):
    """
    Container to store all config attributes.
    """

    def __init__(self, cfg_file):
        super().__init__(cfg_file=cfg_file, section=ConfigSections.PLUGIN)


from anasymod.templates.templ import JinjaTempl
from anasymod.config import EmuConfig
from anasymod.generators.gen_api import SVAPI, ModuleInst
from anasymod.structures.structure_config import StructureConfig
from anasymod.sim_ctrl.datatypes import DigitalCtrlInput, DigitalCtrlOutput, DigitalSignal, AnalogCtrlInput, AnalogCtrlOutput

class ModuleVIOSimCtrl(JinjaTempl):
    def __init__(self, scfg: StructureConfig, plugin_includes: list):
        super().__init__(trim_blocks=True, lstrip_blocks=True)

        ctrl_inputs = scfg.analog_ctrl_inputs + scfg.digital_ctrl_inputs
        ctrl_outputs = scfg.analog_ctrl_outputs + scfg.digital_ctrl_outputs

        #####################################################
        # Add plugin specific includes
        #####################################################

        self.plugin_includes = SVAPI()
        for plugin in plugin_includes:
            for include_statement in plugin.include_statements:
                self.plugin_includes.writeln(f'{include_statement}')

        #####################################################
        # Define module ios
        #####################################################
        self.module_ifc = SVAPI()

        module = ModuleInst(api=self.module_ifc, name="sim_ctrl_gen")
        module.add_inputs(ctrl_outputs)
        module.add_outputs(ctrl_inputs)
        module.add_input(scfg.emu_clk)

        module.generate_header()

        #####################################################
        # PC sim control section
        #####################################################

        ## Custom control IOs for pc sim control module
        self.pc_sim_crtl_ifc = SVAPI()

        # Only generate instantiation if there is any ctrl signal available, that is not a special control signal
        has_custom_ctrl_signal = False
        for ctrl_signal in (ctrl_outputs + ctrl_inputs):
            if ctrl_signal.name not in scfg.special_ctrl_ios:
                has_custom_ctrl_signal = True
                break

        if has_custom_ctrl_signal:
            sim_ctrl_module = ModuleInst(api=self.pc_sim_crtl_ifc, name="sim_ctrl")
            for ctrl_output in ctrl_outputs:
                if ctrl_output.name not in scfg.special_ctrl_ios:
                    sim_ctrl_module.add_input(ctrl_output, connection=ctrl_output)
            for ctrl_input in ctrl_inputs:
                if ctrl_input.name not in scfg.special_ctrl_ios:
                    sim_ctrl_module.add_output(ctrl_input, connection=ctrl_input)

            sim_ctrl_module.generate_instantiation()

        # set number of clk cycles for initial reset
        self.rst_clkcycles = scfg.cfg.rst_clkcycles

        #####################################################
        # FPGA sim control section - Instantiate vio wizard
        #####################################################

        self.vio_wiz_inst = SVAPI()
        vio_wiz = ModuleInst(api=self.vio_wiz_inst, name="vio_0")

        for k, input in enumerate(ctrl_outputs):
            if isinstance(input, AnalogCtrlOutput):
                width = '`LONG_WIDTH_REAL'
            elif isinstance(input, DigitalCtrlOutput):
                width = input.width
            else:
                raise Exception(f"Provided signal type:{type(input)} is not supported!")

            vio_i = DigitalSignal(name=f'probe_in{k}', width=width, abspath=None)
            vio_wiz.add_input(io_obj=vio_i, connection=input)

        for k, output in enumerate(ctrl_inputs):
            if isinstance(output, AnalogCtrlInput):
                width = '`LONG_WIDTH_REAL'
            elif isinstance(output, DigitalCtrlInput):
                width = output.width
            else:
                raise Exception(f"Provided signal type:{type(output)} is not supported!")

            vio_o = DigitalSignal(name=f'probe_out{k}', width=width, abspath=None)
            vio_wiz.add_output(io_obj=vio_o, connection=output)

        vio_wiz.add_input(io_obj=DigitalSignal(name='clk', width=1, abspath=None), connection=scfg.emu_clk)
        vio_wiz.generate_instantiation()

    TEMPLATE_TEXT = '''\
`timescale 1ns/1ps

{{subst.plugin_includes.text}}

`default_nettype none
{{subst.module_ifc.text}}

`ifdef SIMULATION_MODE_MSDSL
    // reset sequence
    logic emu_rst_state = 1'b1;    
    assign emu_rst = emu_rst_state;
    initial begin
        @(posedge emu_clk);
        #(((0.1/(`EMU_CLK_FREQ))*1.0s)+(1.0/(`EMU_CLK_FREQ))*{{subst.rst_clkcycles}}*1.0s);
        emu_rst_state = 1'b0;
    end

    // decimation threshold
    `ifndef DEC_THR_VAL_MSDSL
        `define DEC_THR_VAL_MSDSL 0
    `endif // `ifdef DEC_THR_VAL_MSDSL
    assign emu_dec_thr = `DEC_THR_VAL_MSDSL;

     // stall / run / sleep controls
    logic [((`TIME_WIDTH)-1):0] emu_ctrl_data_state;
    logic [3:0] emu_ctrl_mode_state;
    assign emu_ctrl_data = emu_ctrl_data_state;
    assign emu_ctrl_mode = emu_ctrl_mode_state;

    // module for custom vio handling
    // NOTE: sim_ctrl module must be written and added to the project manually!!!

{{subst.pc_sim_crtl_ifc.text}}
`else
	// VIO instantiation
{{subst.vio_wiz_inst.text}}
`endif // `ifdef SIMULATION_MODE_MSDSL

endmodule

`default_nettype wire
'''

def main():
    print(ModuleVIOSimCtrl(scfg=StructureConfig(prj_cfg=EmuConfig(root='test', cfg_file=''))).render())

if __name__ == "__main__":
    main()


class ModuleBase():
    """
    This is the base class for module structure generation.
    """
    def gen_module(self):
        """
        Generates sv code for module body.
        :return: Generated code of module body.
        """
        pass

    def gen_instantiation(self):
        """
        Generates sv code for module instantiation.
        :return: Generated code of module instantiation.
        """
        pass

    #note: it shall be possible to cluster ports -> all analog clks, all dig clks, all ctrl signals


import os
import os.path
from pathlib import Path
import shutil

# working with paths

def get_full_path(path):
    return str(Path(path).resolve())

def get_sibling(path, sibling):
    return str(Path(path).parent / sibling)

# working with programs

def which(program, path=None):
    return get_full_path(shutil.which(program, path=path))

# working with packages

def anasymod_root():
    return Path(__file__).parent

def anasymod_header():
    return anasymod_root() / 'verilog' / 'anasymod.sv'

def get_from_anasymod(*args):
    return os.path.join(str(anasymod_root()), *args)

# working with directories

def mkdir_p(path):
    Path(path).mkdir(parents=True, exist_ok=True)

def rm_rf(path):
    shutil.rmtree(path, ignore_errors=True)



from anasymod.templates.templ import JinjaTempl
from anasymod.config import EmuConfig
from anasymod.generators.gen_api import SVAPI, ModuleInst
from anasymod.structures.structure_config import StructureConfig
from anasymod.sim_ctrl.datatypes import DigitalSignal


class ModuleClkManager(JinjaTempl):
    """
    This is the generator for clk_gen.sv wrapper.
    """
    def __init__(self, scfg: StructureConfig):
        super().__init__(trim_blocks=True, lstrip_blocks=True)

        #####################################################
        # Create module interface
        #####################################################
        self.module_ifc = SVAPI()

        module = ModuleInst(api=self.module_ifc, name="clk_gen")
        module.add_inputs(scfg.clk_i)
        module.add_output(scfg.dbg_clk)
        module.add_output(scfg.emu_clk_2x)
        module.add_outputs(scfg.clk_independent)
        module.generate_header()

        #####################################################
        # Instantiate clk wizard
        #####################################################
        self.clk_wiz_inst = SVAPI()

        clk_wiz = ModuleInst(api=self.clk_wiz_inst, name='clk_wiz_0')
        clk_wiz.add_inputs(scfg.clk_i, connections=scfg.clk_i)

        # handled by emu clk generator
        for k, port in enumerate([scfg.emu_clk_2x] + [scfg.dbg_clk] + scfg.clk_independent):
            clk_wiz.add_output(DigitalSignal(abspath=None, width=1, name=f'clk_out{k + 1}'), connection=port)

        clk_wiz.add_input(DigitalSignal(abspath=None, width=1, name='reset'), connection=r"1'b0")
        clk_wiz.add_output(DigitalSignal(abspath=None, width=1, name='locked'), DigitalSignal(abspath=None, width=1, name='locked'))

        clk_wiz.generate_instantiation()

        #####################################################
        # Create independent clks for simulation case
        #####################################################

        self.emu_clk = scfg.emu_clk
        self.independent_clks = scfg.clk_independent

    TEMPLATE_TEXT = '''\
`timescale 1ns/1ps

`default_nettype none
{{subst.module_ifc.text}}

`ifdef SIMULATION_MODE_MSDSL
	// emulator clock sequence
	logic emu_clk_2x_state;
{% for clk in subst.independent_clks %}
	logic {{clk.name}}_state;
{% endfor %}
	
	initial begin
		// since the reset signal is initially "1", this delay+posedge will
		// cause the MSDSL templates to be reset
	    emu_clk_2x_state = 1'b0;
	    #((0.25/(`EMU_CLK_FREQ))*1s);
	    emu_clk_2x_state = 1'b1;

	    // clock runs forever
	    forever begin
	        #((0.25/(`EMU_CLK_FREQ))*1s);
	        emu_clk_2x_state = ~emu_clk_2x_state;
	    end
	end

{% for clk in subst.independent_clks %}
    initial begin
        {{clk.name}}_state = 1'b0;
	    forever begin
	        #(({{0.5/clk.freq}})*1s);
	        {{clk.name}}_state = ~{{clk.name}}_state;
	    end
    end
{% endfor %}
	
	// output assignment
	assign emu_clk_2x = emu_clk_2x_state;
{% for clk in subst.independent_clks %}
	assign {{clk.name}} = {{clk.name}}_state;
{% endfor %}	
`else
	logic locked;
    {{subst.clk_wiz_inst.text}}
`endif // `ifdef SIMULATION_MODE_MSDSL

endmodule
`default_nettype wire
'''

def main():
    print(ModuleClkManager(scfg=StructureConfig(prj_cfg=EmuConfig(root='test', cfg_file=''))).render())

if __name__ == "__main__":
    main()


from anasymod.templates.templ import JinjaTempl
from anasymod.targets import FPGATarget

class TemplDbgHub(JinjaTempl):
    def __init__(self, target: FPGATarget):
        super().__init__()
        self.dbg_hub_prop = {}
        self.dbg_hub_prop['C_ENABLE_CLK_DIVIDER'] = 'false'
        self.dbg_hub_prop['C_USER_SCAN_CHAIN'] = '1'
        self.dbg_hub_prop['C_CLK_INPUT_FREQ_HZ'] = str(int(target.prj_cfg.board.dbg_hub_clk_freq))
        self.conn_dbg_clk = f'clk_gen_i/clk_wiz_0_i/clk_out2'
        #ToDo: In case multiple dbg hubs are necessary, managing signal names needs to be adapted and several
        # instantiations are necessary

    TEMPLATE_TEXT = '''
# start auto-generated code for debug hub
{%- for propname, propvalue in subst.dbg_hub_prop.items() %}
set_property {{propname}} {{ propvalue }} [get_debug_cores dbg_hub]
{%- endfor %}
connect_debug_port dbg_hub/clk [get_pins {{ subst.conn_dbg_clk }}]

# end auto-generated code for debug hub
'''

def main():
    print(TemplDbgHub().render())

if __name__ == "__main__":
    main()


import numpy as np

try:
    # import si-prefix from Inicio installation
    from site_pip_packages.si_prefix import si_format
except:
    from si_prefix import si_format

try:
    from vcd import VCDWriter
except:
    print('ERROR: Could not load pyvcd package!')

import datetime

from anasymod.utils.VCD_parser import ParseVCD
from anasymod.enums import ResultFileTypes

class ConvertWaveform():
    """
    Convert raw result files to vcd and also make sure fixed-point datatypes are properly converted to a floating point
    representation. Currently supported raw result datatypes are vcd and csv.
    """
    def __init__(self, str_cfg, result_type_raw, result_path_raw, result_path,
                 float_type=True, emu_time_scaled=True, debug=False,
                 dt_scale=1e-15):
        """

        :param str_cfg: structure config object used in current project.
        :param result_type_raw: filetype of result file to be converted
        :param result_path_raw: path to raw result file
        :param result_path: path to converted result file
        :param float_type: flag to indicate if real signal's data type is fixed-point or floating point
        :param emu_time_scaled: flag to indicate, if signals shall be displayed over cycle count or time
        :param debug: if debug flag is set to true, all signals from result file will be kept, even if they are not a
                        specified probe; keep in mind, that for those signals no fixed to float conversion can be done
        """

        # defaults
        self.result_path_raw = result_path_raw
        scfg = str_cfg
        self.signal_lookup = {}

        # store data from FPGA in a dictionary
        probe_data = {}
        real_signals = set()
        reg_widths = {}

        if result_type_raw == ResultFileTypes.CSV:
            # read CSV file
            with open(self.result_path_raw, 'r') as f:
                first_line = f.readline()

            # split up the first line into comma-delimited names
            signals = first_line.split(',')

            # account for whitespace
            signals = [signal.strip() for signal in signals]

            # strip off the signal indices and add to a lookup table
            for k, signal in enumerate(signals):
                if '[' in signal:
                    signal = signal[:signal.index('[')]
                self.signal_lookup[signal] = k

            # print keys
            print(f'Signals in result file: {[key for key in self.signal_lookup.keys()]}')

            for analog_signal in scfg.analog_probes:
                name = 'trace_port_gen_i/' + analog_signal.name
                if (name) in self.signal_lookup:
                    # add to set of probes with "real" data type
                    real_signals.add(name)

                    # get unscaled data and apply scaling factor
                    probe_data[name] = (2 ** int(analog_signal.exponent)) * self.get_csv_col(name)

                    # convert data to native Python float type (rather than numpy float)
                    # this is required for PyVCD
                    try:
                        probe_data[name] = [float(x) for x in probe_data[name]]
                    except:
                        probe_data[name] = [float(probe_data[name])]

            for digital_signal in scfg.digital_probes + [scfg.dec_cmp] + [scfg.time_probe]:
                name = 'trace_port_gen_i/' + digital_signal.name
                if name in self.signal_lookup:
                    # define width for this probe
                    reg_widths[name] = int(digital_signal.width)

                    # get unscaled data
                    probe_data[name] = self.get_csv_col(name)

                    # convert data to native Python int type (rather than numpy int)
                    # this is required for PyVCD
                    try:
                        probe_data[name] = [int(x) for x in probe_data[name]]
                    except ValueError:
                        print(f'ValueError encountered when converting probe_data[{name}].')
                        print(f'Contents of probe_data[{name}]:')
                        print(f'{probe_data}')
                        print(f'Contents of CSV file {self.result_path_raw}:')
                        print(open(self.result_path_raw, 'r').read())
                        raise

            # Write data to VCD file
            with open(result_path, 'w') as vcd:
                timescale = self.get_pyvcd_timescale(dt_scale)
                with VCDWriter(vcd, timescale=timescale, date=str(datetime.datetime.today())) as writer:
                    # register all of the signals that will be written to VCD
                    reg = {}
                    for sig, scaled_data in probe_data.items():
                        # determine signal scope and name
                        signal_split = sig.split('/')
                        vcd_scope = '.'.join(signal_split[:-1])
                        vcd_name = signal_split[-1]

                        # determine signal type and size
                        if sig in real_signals:
                            vcd_var_type = 'real'
                            vcd_size = None
                        elif sig in reg_widths:
                            vcd_var_type = 'reg'
                            vcd_size = reg_widths[sig]
                        else:
                            raise Exception('Unknown signal type.')

                        # register the signal
                        reg[sig] = writer.register_var(scope=vcd_scope, name=vcd_name,
                                                                    var_type=vcd_var_type,
                                                                    size=vcd_size)

                    # iterate over all timesteps
                    prev_timestep = float('-inf')
                    for k, timestamp in enumerate(probe_data['trace_port_gen_i/' + scfg.time_probe.name]):
                        # break if the current timestep is less than the previous one, since that means
                        # wrapping has occurred.  (only if emu_time_scaled is True, meaning that
                        # we're using emu_time on the x axis, not cycle number)
                        if emu_time_scaled and (timestamp < prev_timestep):
                            break
                        prev_timestep = timestamp

                        # Set the x-axis value according to emu_time_scaled: True means use
                        # emu_time, and False means use the cycle number
                        if emu_time_scaled:
                            chg_val = timestamp
                        else:
                            chg_val = k

                        # iterate over all signals and log their change at this timestamp
                        for sig, scaled_data in probe_data.items():
                            writer.change(reg[sig], chg_val, scaled_data[k])

        elif result_type_raw == ResultFileTypes.VCD:
            vcd_file_name = result_path_raw
            vcd_handle = ParseVCD(vcd_file_name)
            signal_dict = vcd_handle.parse_vcd(update_data=False)

            # print signal names
            signal_names = [(signal_dict[key]["nets"][0]["hier"] + '.' + signal_dict[key]["nets"][0]["name"], key) for key in signal_dict.keys()]
            print(f'Signals in result file: {[sig_name[0] for sig_name in signal_names]}')

            for analog_signal in scfg.analog_probes:
                analog_signal_path = 'top.trace_port_gen_i' + '.' + analog_signal.name
                if analog_signal_path in [sig[0] for sig in signal_names]:
                    # add to set of probes with "real" data type
                    real_signals.add(analog_signal_path)

                    probe_data[analog_signal_path] = {}
                    probe_data[analog_signal_path]['index'] = 0

                    # get the signal identifier from analog_signal used in VCD file
                    signal_identifier = signal_names[[y[0] for y in signal_names].index(analog_signal_path)][1]
                    data = []
                    for c,v in signal_dict[signal_identifier]['cv']:
                        if not isinstance(v, float):
                            for k in range(analog_signal.width - len(v)): # extend binary to full width, some simulators don't do that by default and remove leading zeros in VCD file
                                v = '0' + v
                            if v[0] == '1': # check if number is negative and if so calculate the two's complement
                                v = -1 * (int((''.join('1' if x == '0' else '0' for x in v)), 2) + 1)
                            else:
                                v = int(v, 2)
                        if not float_type:
                            v = 2 ** int(analog_signal.exponent) * v
                        data.append((c, v))
                    probe_data[analog_signal_path]['data'] = np.asarray(data)

                    # convert data to native Python float type (rather than numpy float) this is required for PyVCD
                    probe_data[analog_signal_path]['data'] = [(int(c), float(v)) for c, v in probe_data[analog_signal_path]['data']]

            for digital_signal in scfg.digital_probes + [scfg.dec_cmp] + [scfg.time_probe]:
                digital_signal_path = 'top.trace_port_gen_i' + '.' + digital_signal.name
                if digital_signal_path in [sig[0] for sig in signal_names]:
                    # define width for this probe
                    reg_widths[digital_signal_path] = int(digital_signal.width)

                    probe_data[digital_signal_path] = {}
                    probe_data[digital_signal_path]['index'] = 0

                    # get unscaled data
                    probe_data[digital_signal_path]['data'] = np.asarray([(c, v) for c,v in signal_dict[signal_names[[y[0] for y in signal_names].index(digital_signal_path)][1]]['cv']])

                    # convert data to native Python int type (rather than numpy int) this is required for PyVCD
                    data = []
                    for c, v in probe_data[digital_signal_path]['data']:
                        try:
                            data.append((int(c), int(v, 2)))
                        except: # In case of an x or z value, a 0 will be added; this is necessary for PYVCD
                            data.append((int(c), int(0)))

                    probe_data[digital_signal_path]['data'] = data

            # Write data to VCD file

            with open(result_path, 'w') as vcd:
                timescale = self.get_pyvcd_timescale(dt_scale)
                with VCDWriter(vcd, timescale=timescale, date=str(datetime.datetime.today())) as writer:
                    # register all of the signals that will be written to VCD
                    reg = {}
                    for sig, scaled_data in probe_data.items():
                        # determine signal scope and name
                        signal_split = sig.split('.')
                        vcd_scope = '.'.join(signal_split[:-1])
                        vcd_name = signal_split[-1]

                        # determine signal type and size
                        if sig in real_signals:
                            vcd_var_type = 'real'
                            vcd_size = None
                        elif sig in reg_widths:
                            vcd_var_type = 'reg'
                            vcd_size = reg_widths[sig]
                        else:
                            raise Exception('Unknown signal type.')

                        # register the signal
                        reg[sig] = writer.register_var(scope=vcd_scope, name=vcd_name,
                                                                    var_type=vcd_var_type,
                                                                    size=vcd_size)

                    prev_timestep = float('-inf')
                    # Add all other signals in case debug flag is set
                    if debug:
                        for signal in signal_names:
                            if not signal[0] in probe_data.keys(): # signal was not listed yet
                                var_type = signal_dict[signal[1]]['nets'][0]['type']
                                if var_type != 'parameter':
                                    name = signal_dict[signal[1]]['nets'][0]['name']
                                    scope = signal_dict[signal[1]]['nets'][0]['hier']
                                    path = scope + '.' + name
                                    size = signal_dict[signal[1]]['nets'][0]['size']

                                    # register the signal
                                    reg[path] = writer.register_var(scope=scope, name=name, var_type=var_type, size=int(size))

                    # time probe path
                    time_path = 'top.trace_port_gen_i' + '.' + scfg.time_probe.name

                    # calculate emu_time offset
                    offset = 0
                    for idx, (cycle_count, timestamp) in enumerate(probe_data[time_path]['data']):
                        # break if the current timestep is less than the previous one, since that means
                        # wrapping has occurred
                        if timestamp < prev_timestep:
                            break
                        prev_timestep = timestamp

                        if timestamp > 0:
                            break

                        offset = cycle_count

                    #############################
                    # Represent signals over time
                    #############################

                    if emu_time_scaled:
                        # Add an index to all signals in signal dict, in case debug option was selected:
                        if debug:
                            for sig in signal_dict.keys():
                                signal_dict[sig]['index'] = 0

                        for idx, (cycle_count, timestamp) in enumerate(probe_data[time_path]['data']):
                            # break if timestamp is less than zero since it means that wrapping has occurred
                            if timestamp < 0:
                                break

                            # store all available signals in list, any signal, that is no longer in the list will be skipped
                            # signals, where the cycle_count is nop longer between current and next cycle count of the time
                            # signal will be removed from the list
                            sigs_in_interval = []
                            for sig in probe_data.keys():
                                sigs_in_interval.append((sig, None))

                            # Add all other signals in case debug flag is set
                            if debug:
                                for signal in signal_names:
                                    if not signal[0] in probe_data.keys() and not signal_dict[signal[1]]['nets'][0]['type'] == 'parameter':  # signal was not listed yet
                                        sigs_in_interval.append((signal[0], signal[1]))

                            # list of all activities within time interval
                            timestep_events = []

                            # As soon as all signals are no longer in the interval, timestep will advance
                            while sigs_in_interval:
                                for sig in sigs_in_interval:
                                    if sig[0] in probe_data.keys():
                                        sig_tuple = probe_data[sig[0]]['data'][probe_data[sig[0]]['index']]
                                    else: # debug signals
                                        sig_tuple = signal_dict[sig[1]]['cv'][signal_dict[sig[1]]['index']]

                                    if cycle_count == sig_tuple[0]:
                                        timestep_events.append([sig, timestamp, sig_tuple[1]])

                                        if sig[0] in probe_data.keys():
                                            probe_data[sig[0]]['index'] += 1
                                        else:
                                            signal_dict[sig[1]]['index'] += 1

                                        # This signal no longer needs to be observed
                                        sigs_in_interval.remove(sig)
                                    else:
                                        try: # Check if time signal is already at the end
                                            next_timestamp = probe_data[time_path]['data'][idx + 1][0]
                                        except: # Time signal has finished the end, finish conversion
                                            sigs_in_interval = []
                                            break
                                        if round(next_timestamp - offset) > sig_tuple[0]:
                                            # Note: There is always an offset between cycle count and timestamp  -> substract offset
                                            # The cycle count from data signal does not have a match with emu_time signal's cycle count
                                            # -> we need to apply interpolation in order to assign the timestamp properly
                                            cycles_in_dt = probe_data[time_path]['data'][idx+1][0] - cycle_count
                                            dt = probe_data[time_path]['data'][idx+1][1] - timestamp
                                            interp_timestamp = int(dt/cycles_in_dt * (sig_tuple[0] - cycle_count + offset) + timestamp)

                                            timestep_events.append([sig, interp_timestamp, sig_tuple[1]])

                                            if sig[0] in probe_data.keys():
                                                probe_data[sig[0]]['index'] += 1
                                            else:
                                                signal_dict[sig[1]]['index'] += 1
                                        else:
                                            # This signal no longer needs to be observed
                                            sigs_in_interval.remove(sig)

                            # Register events in time interval in chronological order
                            timestep_events = sorted(timestep_events, key=self.sort_timestamp)
                            for [name, timestamp, value] in timestep_events:
                                writer.change(reg[name[0]], timestamp, value)

                    ####################################
                    # Represent signals over cycle count
                    ####################################

                    else:
                        time_events = []  # store all events from result file
                        for signal in probe_data.keys():
                            for sig_tuple in probe_data[signal]['data']:
                                time_events.append([signal, sig_tuple[0], sig_tuple[1]])

                        # Add all other signals in case debug flag is set
                        if debug:
                            for signal in signal_names:
                                if not signal[0] in probe_data.keys() and not signal_dict[signal[1]]['nets'][0]['type'] == 'parameter':  # signal was not listed yet
                                    name = signal_dict[signal[1]]['nets'][0]['name']
                                    hier = signal_dict[signal[1]]['nets'][0]['hier']
                                    path = hier + '.' + name
                                    for sig_tuple in signal_dict[signal[1]]['cv']:
                                        time_events.append([path, sig_tuple[0], sig_tuple[1]])

                        # Register events in chronological order
                        time_events = sorted(time_events, key=self.sort_timestamp)
                        for [sig_name, timestamp, value] in time_events:
                            writer.change(reg[sig_name], timestamp, value)



        else:
            raise Exception(f'ERROR: No supported Result file format selected:{result_type_raw}')

    def get_csv_col(self, name):
        """
        Getting unscaled data from csv file column
        :return:
        """
        # have to preview the first few lines
        with open(self.result_path_raw, 'r') as f:
            first_line = f.readline()
            second_line = f.readline()

        # determine how many lines to skip
        if second_line.startswith('Radix'):
            skip_header=2
        else:
            skip_header=1

        # call a NumPy routine to read from the CSV file
        return np.genfromtxt(
            self.result_path_raw,
            delimiter=',',
            usecols=self.signal_lookup[name],
            skip_header=skip_header
        )

    def sort_timestamp(self, element):
        return element[1]

    def get_pyvcd_timescale(self, val):
        return si_format(val, precision=0) + 's'